#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j.io._bolt4 import Bolt4x1
from neo4j.conf import PoolConfig


def test_conn_timed_out(fake_socket):
    address = ("127.0.0.1", 7687)
    max_connection_lifetime = 0
    connection = Bolt4x1(address, fake_socket(address), max_connection_lifetime)
    assert connection.timedout() is True


def test_conn_not_timed_out_if_not_enabled(fake_socket):
    address = ("127.0.0.1", 7687)
    max_connection_lifetime = -1
    connection = Bolt4x1(address, fake_socket(address), max_connection_lifetime)
    assert connection.timedout() is False


def test_conn_not_timed_out(fake_socket):
    address = ("127.0.0.1", 7687)
    max_connection_lifetime = 999999999
    connection = Bolt4x1(address, fake_socket(address), max_connection_lifetime)
    assert connection.timedout() is False


def test_db_extra_in_begin(fake_socket):
    address = ("127.0.0.1", 7687)
    socket = fake_socket(address)
    connection = Bolt4x1(address, socket, PoolConfig.max_connection_lifetime)
    connection.begin(db="something")
    connection.send_all()
    tag, fields = socket.pop_message()
    assert tag == b"\x11"
    assert len(fields) == 1
    assert fields[0] == {"db": "something"}


def test_db_extra_in_run(fake_socket):
    address = ("127.0.0.1", 7687)
    socket = fake_socket(address)
    connection = Bolt4x1(address, socket, PoolConfig.max_connection_lifetime)
    connection.run("", {}, db="something")
    connection.send_all()
    tag, fields = socket.pop_message()
    assert tag == b"\x10"
    assert len(fields) == 3
    assert fields[0] == ""
    assert fields[1] == {}
    assert fields[2] == {"db": "something"}


def test_n_extra_in_discard(fake_socket):
    address = ("127.0.0.1", 7687)
    socket = fake_socket(address)
    connection = Bolt4x1(address, socket, PoolConfig.max_connection_lifetime)
    connection.discard(n=666)
    connection.send_all()
    tag, fields = socket.pop_message()
    assert tag == b"\x2F"
    assert len(fields) == 1
    assert fields[0] == {"n": 666}


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (666, {"n": -1, "qid": 666}),
        (-1, {"n": -1}),
    ]
)
def test_qid_extra_in_discard(fake_socket, test_input, expected):
    address = ("127.0.0.1", 7687)
    socket = fake_socket(address)
    connection = Bolt4x1(address, socket, PoolConfig.max_connection_lifetime)
    connection.discard(qid=test_input)
    connection.send_all()
    tag, fields = socket.pop_message()
    assert tag == b"\x2F"
    assert len(fields) == 1
    assert fields[0] == expected


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (777, {"n": 666, "qid": 777}),
        (-1, {"n": 666}),
    ]
)
def test_n_and_qid_extras_in_discard(fake_socket, test_input, expected):
    # python -m pytest tests/unit/io/test_class_bolt4x0.py -s -k test_n_and_qid_extras_in_discard
    address = ("127.0.0.1", 7687)
    socket = fake_socket(address)
    connection = Bolt4x1(address, socket, PoolConfig.max_connection_lifetime)
    connection.discard(n=666, qid=test_input)
    connection.send_all()
    tag, fields = socket.pop_message()
    assert tag == b"\x2F"
    assert len(fields) == 1
    assert fields[0] == expected


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (666, {"n": 666}),
        (-1, {"n": -1}),
    ]
)
def test_n_extra_in_pull(fake_socket, test_input, expected):
    address = ("127.0.0.1", 7687)
    socket = fake_socket(address)
    connection = Bolt4x1(address, socket, PoolConfig.max_connection_lifetime)
    connection.pull(n=test_input)
    connection.send_all()
    tag, fields = socket.pop_message()
    assert tag == b"\x3F"
    assert len(fields) == 1
    assert fields[0] == expected


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (777, {"n": -1, "qid": 777}),
        (-1, {"n": -1}),
    ]
)
def test_qid_extra_in_pull(fake_socket, test_input, expected):
    # python -m pytest tests/unit/io/test_class_bolt4x0.py -s -k test_qid_extra_in_pull
    address = ("127.0.0.1", 7687)
    socket = fake_socket(address)
    connection = Bolt4x1(address, socket, PoolConfig.max_connection_lifetime)
    connection.pull(qid=test_input)
    connection.send_all()
    tag, fields = socket.pop_message()
    assert tag == b"\x3F"
    assert len(fields) == 1
    assert fields[0] == expected


def test_n_and_qid_extras_in_pull(fake_socket):
    address = ("127.0.0.1", 7687)
    socket = fake_socket(address)
    connection = Bolt4x1(address, socket, PoolConfig.max_connection_lifetime)
    connection.pull(n=666, qid=777)
    connection.send_all()
    tag, fields = socket.pop_message()
    assert tag == b"\x3F"
    assert len(fields) == 1
    assert fields[0] == {"n": 666, "qid": 777}


def test_hello_passes_routing_metadata(fake_socket_pair):
    address = ("127.0.0.1", 7687)
    sockets = fake_socket_pair(address)
    # TODO helper method for encoding messages
    sockets.server.sendall(b"\x00\x03\xB1\x70\xA0\x00\x00")
    connection = Bolt4x1(address, sockets.client, PoolConfig.max_connection_lifetime,
                         routing_context={"foo": "bar"})
    connection.hello()
    tag, fields = sockets.server.pop_message()
    assert tag == 0x01
    assert len(fields) == 1
    assert fields[0]["routing"] == {"foo": "bar"}



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import subprocess
import os
import time

from platform import system
from threading import Thread
from time import sleep

from boltkit.server.stub import BoltStubService
from pytest import fixture

import logging
log = logging.getLogger("neo4j")

# from neo4j.debug import watch
# watch("neo4j")


class StubServer:

    def __init__(self, port, script):
        self.port = port
        self.script = os.path.join(os.path.dirname(__file__), "scripts", script)

    def run(self):
        shell = system() == "Windows"  # I hate myself for doing this
        self._process = subprocess.Popen(["python", "-m", "boltkit", "stub", "-v", "-l", ":{}".format(str(self.port)), "-t", "10", self.script], stdout=subprocess.PIPE, shell=shell)
        # Need verbose for this to work
        line = self._process.stdout.readline().decode("utf-8")
        log.debug("started stub server {}".format(self.port))
        log.debug(line.strip("\n"))

    def wait(self):
        while True:
            return_code = self._process.poll()
            if return_code is not None:
                line = self._process.stdout.readline().decode("utf-8")
                if line == "":
                    break
                log.debug(line.strip("\n"))

        return True

    def kill(self):
        # Kill process if not already dead
        if self._process.poll() is None:
            self._process.kill()


class StubCluster:

    def __init__(self, servers):
        self.servers = {port: StubServer(port, script) for port, script in dict(servers).items()}

    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_value, traceback):
        self.wait()

    def start(self):
        for port, server in self.servers.items():
            server.run()

    def wait(self):
        success = True
        for port, server in self.servers.items():
            if not server.wait():
                success = False
            server.kill()

        if not success:
            raise Exception("Stub server failed")


class LegacyStubServer(Thread):

    def __init__(self, port, script):
        super(LegacyStubServer, self).__init__()
        self.port = port
        self.script = os.path.join(os.path.dirname(__file__), "scripts", script)

    def run(self):
        check_call(["python", "-m", "boltkit.legacy.stub", "-v", str(self.port), self.script])


class LegacyStubCluster:

    def __init__(self, servers):
        self.servers = {port: LegacyStubServer(port, script) for port, script in dict(servers).items()}

    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_value, traceback):
        self.wait()

    def start(self):
        for port, server in self.servers.items():
            server.start()
        sleep(0.5)

    def wait(self):
        for port, server in self.servers.items():
            server.join()


class DefaultBoltStubService(BoltStubService):

    default_base_port = 9001


class StubCluster(StubCluster):

    def __init__(self, *servers):
        print("")
        scripts = [os.path.join(os.path.dirname(__file__), "scripts", server) for server in servers]

        bss = DefaultBoltStubService.load(*scripts)
        servers2 = {port: script.filename for port, script in bss.scripts.items()}
        super().__init__(servers2)

        # def run():
        #     check_call(["bolt", "stub", "-v", "-t", "10", "-l", ":9001"] + scripts)

        # self.thread = Thread(target=run)

    # def __enter__(self):
    #     self.thread.start()
    #     sleep(0.5)

    # def __exit__(self, exc_type, exc_value, traceback):
    #     self.thread.join(3)


@fixture
def script():
    return lambda *paths: path_join(dirname(__file__), "scripts", *paths)


@fixture
def driver_info():
    """ Base class for test cases that integrate with a server.
    """
    return {
        "uri_bolt": "bolt://localhost:9001",
        "uri_neo4j": "neo4j://localhost:9001",
        "user": "test",
        "password": "test",
        "auth_token": ("test", "test")
    }



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import (
    time,
    datetime,
    timedelta,
)

from neo4j.packstream import Structure
from neo4j.time import (
    Duration,
    Date,
    Time,
    DateTime,
)


def get_date_unix_epoch():
    return Date(1970, 1, 1)


def get_date_unix_epoch_ordinal():
    return get_date_unix_epoch().to_ordinal()


def get_datetime_unix_epoch_utc():
    from pytz import utc
    return DateTime(1970, 1, 1, 0, 0, 0, utc)


def hydrate_date(days):
    """ Hydrator for `Date` values.

    :param days:
    :return: Date
    """
    return Date.from_ordinal(get_date_unix_epoch_ordinal() + days)


def dehydrate_date(value):
    """ Dehydrator for `date` values.

    :param value:
    :type value: Date
    :return:
    """
    return Structure(b"D", value.toordinal() - get_date_unix_epoch().toordinal())


def hydrate_time(nanoseconds, tz=None):
    """ Hydrator for `Time` and `LocalTime` values.

    :param nanoseconds:
    :param tz:
    :return: Time
    """
    from pytz import FixedOffset
    seconds, nanoseconds = map(int, divmod(nanoseconds, 1000000000))
    minutes, seconds = map(int, divmod(seconds, 60))
    hours, minutes = map(int, divmod(minutes, 60))
    seconds = (1000000000 * seconds + nanoseconds) / 1000000000
    t = Time(hours, minutes, seconds)
    if tz is None:
        return t
    tz_offset_minutes, tz_offset_seconds = divmod(tz, 60)
    zone = FixedOffset(tz_offset_minutes)
    return zone.localize(t)


def dehydrate_time(value):
    """ Dehydrator for `time` values.

    :param value:
    :type value: Time
    :return:
    """
    if isinstance(value, Time):
        nanoseconds = int(value.ticks * 1000000000)
    elif isinstance(value, time):
        nanoseconds = (3600000000000 * value.hour + 60000000000 * value.minute +
                       1000000000 * value.second + 1000 * value.microsecond)
    else:
        raise TypeError("Value must be a neo4j.time.Time or a datetime.time")
    if value.tzinfo:
        return Structure(b"T", nanoseconds, value.tzinfo.utcoffset(value).seconds)
    else:
        return Structure(b"t", nanoseconds)


def hydrate_datetime(seconds, nanoseconds, tz=None):
    """ Hydrator for `DateTime` and `LocalDateTime` values.

    :param seconds:
    :param nanoseconds:
    :param tz:
    :return: datetime
    """
    from pytz import FixedOffset, timezone
    minutes, seconds = map(int, divmod(seconds, 60))
    hours, minutes = map(int, divmod(minutes, 60))
    days, hours = map(int, divmod(hours, 24))
    seconds = (1000000000 * seconds + nanoseconds) / 1000000000
    t = DateTime.combine(Date.from_ordinal(get_date_unix_epoch_ordinal() + days), Time(hours, minutes, seconds))
    if tz is None:
        return t
    if isinstance(tz, int):
        tz_offset_minutes, tz_offset_seconds = divmod(tz, 60)
        zone = FixedOffset(tz_offset_minutes)
    else:
        zone = timezone(tz)
    return zone.localize(t)


def dehydrate_datetime(value):
    """ Dehydrator for `datetime` values.

    :param value:
    :type value: datetime
    :return:
    """

    def seconds_and_nanoseconds(dt):
        if isinstance(dt, datetime):
            dt = DateTime.from_native(dt)
        zone_epoch = DateTime(1970, 1, 1, tzinfo=dt.tzinfo)
        dt_clock_time = dt.to_clock_time()
        zone_epoch_clock_time = zone_epoch.to_clock_time()
        t = dt_clock_time - zone_epoch_clock_time
        return t.seconds, t.nanoseconds

    tz = value.tzinfo
    if tz is None:
        # without time zone
        from pytz import utc
        value = utc.localize(value)
        seconds, nanoseconds = seconds_and_nanoseconds(value)
        return Structure(b"d", seconds, nanoseconds)
    elif hasattr(tz, "zone") and tz.zone:
        # with named time zone
        seconds, nanoseconds = seconds_and_nanoseconds(value)
        return Structure(b"f", seconds, nanoseconds, tz.zone)
    else:
        # with time offset
        seconds, nanoseconds = seconds_and_nanoseconds(value)
        return Structure(b"F", seconds, nanoseconds, tz.utcoffset(value).seconds)


def hydrate_duration(months, days, seconds, nanoseconds):
    """ Hydrator for `Duration` values.

    :param months:
    :param days:
    :param seconds:
    :param nanoseconds:
    :return: `duration` namedtuple
    """
    return Duration(months=months, days=days, seconds=seconds, nanoseconds=nanoseconds)


def dehydrate_duration(value):
    """ Dehydrator for `duration` values.

    :param value:
    :type value: Duration
    :return:
    """
    return Structure(b"E", value.months, value.days, value.seconds, int(1000000000 * value.subseconds))


def dehydrate_timedelta(value):
    """ Dehydrator for `timedelta` values.

    :param value:
    :type value: timedelta
    :return:
    """
    months = 0
    days = value.days
    seconds = value.seconds
    nanoseconds = 1000 * value.microseconds
    return Structure(b"E", months, days, seconds, nanoseconds)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j.work.simple import Query
from neo4j.exceptions import Neo4jError, ClientError, TransientError
from neo4j.graph import Node, Relationship
from neo4j.api import Version


def test_can_run_simple_statement(session):
    result = session.run("RETURN 1 AS n")
    for record in result:
        assert record[0] == 1
        assert record["n"] == 1
        with pytest.raises(KeyError):
            _ = record["x"]
        assert record["n"] == 1
        with pytest.raises(KeyError):
            _ = record["x"]
        with pytest.raises(TypeError):
            _ = record[object()]
        assert repr(record)
        assert len(record) == 1


def test_can_run_simple_statement_with_params(session):
    count = 0
    for record in session.run("RETURN $x AS n",
                              {"x": {"abc": ["d", "e", "f"]}}):
        assert record[0] == {"abc": ["d", "e", "f"]}
        assert record["n"] == {"abc": ["d", "e", "f"]}
        assert repr(record)
        assert len(record) == 1
        count += 1
    assert count == 1


@pytest.mark.skip(reason="BOOKMARK, AttributeError: 'Session' object has no attribute 'last_bookmark'")
def test_autocommit_transactions_use_bookmarks(neo4j_driver):
    bookmarks = []
    # Generate an initial bookmark
    with neo4j_driver.session() as session:
        session.run("CREATE ()").consume()
        bookmark = session.last_bookmark()
        assert bookmark is not None
        bookmarks.append(bookmark)
    # Propagate into another session
    with neo4j_driver.session(bookmarks=bookmarks) as session:
        assert list(session.next_bookmarks()) == bookmarks
        session.run("CREATE ()").consume()
        bookmark = session.last_bookmark()
        assert bookmark is not None
        assert bookmark not in bookmarks


def test_fails_on_bad_syntax(session):
    with pytest.raises(Neo4jError):
        session.run("X").consume()


def test_fails_on_missing_parameter(session):
    with pytest.raises(Neo4jError):
        session.run("RETURN {x}").consume()


def test_keys_with_an_error(session):
    with pytest.raises(Neo4jError):
        result = session.run("X")
        list(result.keys())


def test_should_not_allow_empty_statements(session):
    with pytest.raises(ValueError):
        _ = session.run("")


def test_can_run_statement_that_returns_multiple_records(session):
    count = 0
    for record in session.run("unwind(range(1, 10)) AS z RETURN z"):
        assert 1 <= record[0] <= 10
        count += 1
    assert count == 10


def test_can_use_with_to_auto_close_session(session):
    record_list = list(session.run("RETURN 1"))
    assert len(record_list) == 1
    for record in record_list:
        assert record[0] == 1


def test_can_return_node(neo4j_driver):
    with neo4j_driver.session() as session:
        record_list = list(session.run("CREATE (a:Person {name:'Alice'}) "
                                       "RETURN a"))
        assert len(record_list) == 1
        for record in record_list:
            alice = record[0]
            assert isinstance(alice, Node)
            assert alice.labels == {"Person"}
            assert dict(alice) == {"name": "Alice"}


def test_can_return_relationship(neo4j_driver):
    with neo4j_driver.session() as session:
        record_list = list(session.run("CREATE ()-[r:KNOWS {since:1999}]->() "
                                       "RETURN r"))
        assert len(record_list) == 1
        for record in record_list:
            rel = record[0]
            assert isinstance(rel, Relationship)
            assert rel.type == "KNOWS"
            assert dict(rel) == {"since": 1999}


# TODO: re-enable after server bug is fixed
# def test_can_return_path(session):
#     with self.driver.session() as session:
#         record_list = list(session.run("MERGE p=({name:'Alice'})-[:KNOWS]->"
#                                        "({name:'Bob'}) RETURN p"))
#         assert len(record_list) == 1
#         for record in record_list:
#             path = record[0]
#             assert isinstance(path, Path)
#             assert path.start_node["name"] == "Alice"
#             assert path.end_node["name"] == "Bob"
#             assert path.relationships[0].type == "KNOWS"
#             assert len(path.nodes) == 2
#             assert len(path.relationships) == 1


def test_keys_are_available_before_and_after_stream(session):
    result = session.run("UNWIND range(1, 10) AS n RETURN n")
    assert list(result.keys()) == ["n"]
    list(result)
    assert list(result.keys()) == ["n"]


def test_result_single_record_value(session):
    record = session.run(Query("RETURN $x"), x=1).single()
    assert record.value() == 1


@pytest.mark.parametrize(
    "test_input, neo4j_version",
    [
        ("CALL dbms.getTXMetaData", Version(3, 0)),
        ("CALL tx.getMetaData", Version(4, 0)),
    ]
)
def test_autocommit_transactions_should_support_metadata(session, test_input, neo4j_version):
    # python -m pytest tests/integration/test_autocommit.py -s -r fEsxX -k test_autocommit_transactions_should_support_metadata
    metadata_in = {"foo": "bar"}

    result = session.run("RETURN 1")
    value = result.single().value()
    summary = result.consume()
    server_agent = summary.server.agent

    try:
        statement = Query(test_input, metadata=metadata_in)
        result = session.run(statement)
        metadata_out = result.single().value()
    except ClientError as e:
        if e.code == "Neo.ClientError.Procedure.ProcedureNotFound":
            pytest.skip("Cannot assert correct metadata as {} does not support procedure '{}' introduced in Neo4j {}".format(server_agent, test_input, neo4j_version))
        else:
            raise
    else:
        assert metadata_in == metadata_out


def test_autocommit_transactions_should_support_timeout(neo4j_driver):
    with neo4j_driver.session() as s1:
        s1.run("CREATE (a:Node)").consume()
        with neo4j_driver.session() as s2:
            tx1 = s1.begin_transaction()
            tx1.run("MATCH (a:Node) SET a.property = 1").consume()
            try:
                result = s2.run(Query("MATCH (a:Node) SET a.property = 2", timeout=0.25))
                result.consume()
            # On 4.0 and older
            except TransientError:
                pass
            # On 4.1 and forward
            except ClientError:
                pass
            else:
                raise


def test_regex_in_parameter(session):
    matches = []
    result = session.run("UNWIND ['A', 'B', 'C', 'A B', 'B C', 'A B C', "
                          "'A BC', 'AB C'] AS t WITH t "
                          "WHERE t =~ $re RETURN t", re=r'.*\bB\b.*')
    for record in result:
        matches.append(record.value())
    assert matches == ["B", "A B", "B C", "A B C"]


def test_regex_inline(session):
    matches = []
    result = session.run(r"UNWIND ['A', 'B', 'C', 'A B', 'B C', 'A B C', "
                          r"'A BC', 'AB C'] AS t WITH t "
                          r"WHERE t =~ '.*\\bB\\b.*' RETURN t")
    for record in result:
        matches.append(record.value())
    assert matches == ["B", "A B", "B C", "A B C"]


def test_automatic_reset_after_failure(session):
    try:
        result = session.run("X")
        result.consume()
    except Neo4jError:
        result = session.run("RETURN 1")
        record = next(iter(result))
        assert record[0] == 1
    else:
        assert False, "A Cypher error should have occurred"


def test_large_values(bolt_driver):
    for i in range(1, 7):
        with bolt_driver.session() as session:
            session.run("RETURN '{}'".format("A" * 2 ** 20))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from neo4j.io import Bolt

# python -m pytest tests/unit/io/test_class_bolt.py -s -v


def test_class_method_protocol_handlers():
    # python -m pytest tests/unit/io/test_class_bolt.py -s -v -k test_class_method_protocol_handlers
    protocol_handlers = Bolt.protocol_handlers()
    assert len(protocol_handlers) == 4


@pytest.mark.parametrize(
    "test_input, expected",
    [
        ((0, 0), 0),
        ((4, 0), 1),
    ]
)
def test_class_method_protocol_handlers_with_protocol_version(test_input, expected):
    # python -m pytest tests/unit/io/test_class_bolt.py -s -v -k test_class_method_protocol_handlers_with_protocol_version
    protocol_handlers = Bolt.protocol_handlers(protocol_version=test_input)
    assert len(protocol_handlers) == expected


def test_class_method_protocol_handlers_with_invalid_protocol_version():
    # python -m pytest tests/unit/io/test_class_bolt.py -s -v -k test_class_method_protocol_handlers_with_invalid_protocol_version
    with pytest.raises(TypeError):
        Bolt.protocol_handlers(protocol_version=2)


def test_class_method_get_handshake():
    # python -m pytest tests/unit/io/test_class_bolt.py -s -v -k test_class_method_get_handshake
    handshake = Bolt.get_handshake()
    assert handshake == b"\x00\x00\x02\x04\x00\x00\x01\x04\x00\x00\x00\x04\x00\x00\x00\x03"


def test_magic_preamble():
    # python -m pytest tests/unit/io/test_class_bolt.py -s -v -k test_magic_preamble
    preamble = 0x6060B017
    preamble_bytes = preamble.to_bytes(4, byteorder="big")
    assert Bolt.MAGIC_PREAMBLE == preamble_bytes



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import struct
from collections import OrderedDict
from io import BytesIO
from math import pi
from unittest import TestCase
from uuid import uuid4

from pytest import raises

from neo4j.packstream import Packer, UnpackableBuffer, Unpacker, Structure


class PackStreamTestCase(TestCase):

    @classmethod
    def packb(cls, *values):
        stream = BytesIO()
        packer = Packer(stream)
        for value in values:
            packer.pack(value)
        return stream.getvalue()

    @classmethod
    def assert_packable(cls, value, packed_value):
        stream_out = BytesIO()
        packer = Packer(stream_out)
        packer.pack(value)
        packed = stream_out.getvalue()
        try:
            assert packed == packed_value
        except AssertionError:
            raise AssertionError("Packed value %r is %r instead of expected %r" %
                                 (value, packed, packed_value))
        unpacked = Unpacker(UnpackableBuffer(packed)).unpack()
        try:
            assert unpacked == value
        except AssertionError:
            raise AssertionError("Unpacked value %r is not equal to original %r" % (unpacked, value))

    def test_none(self):
        self.assert_packable(None, b"\xC0")

    def test_boolean(self):
        self.assert_packable(True, b"\xC3")
        self.assert_packable(False, b"\xC2")

    def test_negative_tiny_int(self):
        for z in range(-16, 0):
            self.assert_packable(z, bytes(bytearray([z + 0x100])))

    def test_positive_tiny_int(self):
        for z in range(0, 128):
            self.assert_packable(z, bytes(bytearray([z])))

    def test_negative_int8(self):
        for z in range(-128, -16):
            self.assert_packable(z, bytes(bytearray([0xC8, z + 0x100])))

    def test_positive_int16(self):
        for z in range(128, 32768):
            expected = b"\xC9" + struct.pack(">h", z)
            self.assert_packable(z, expected)

    def test_negative_int16(self):
        for z in range(-32768, -128):
            expected = b"\xC9" + struct.pack(">h", z)
            self.assert_packable(z, expected)

    def test_positive_int32(self):
        for e in range(15, 31):
            z = 2 ** e
            expected = b"\xCA" + struct.pack(">i", z)
            self.assert_packable(z, expected)

    def test_negative_int32(self):
        for e in range(15, 31):
            z = -(2 ** e + 1)
            expected = b"\xCA" + struct.pack(">i", z)
            self.assert_packable(z, expected)

    def test_positive_int64(self):
        for e in range(31, 63):
            z = 2 ** e
            expected = b"\xCB" + struct.pack(">q", z)
            self.assert_packable(z, expected)

    def test_negative_int64(self):
        for e in range(31, 63):
            z = -(2 ** e + 1)
            expected = b"\xCB" + struct.pack(">q", z)
            self.assert_packable(z, expected)

    def test_integer_positive_overflow(self):
        with raises(OverflowError):
            self.packb(2 ** 63 + 1)

    def test_integer_negative_overflow(self):
        with raises(OverflowError):
            self.packb(-(2 ** 63) - 1)

    def test_zero_float64(self):
        zero = 0.0
        expected = b"\xC1" + struct.pack(">d", zero)
        self.assert_packable(zero, expected)

    def test_tau_float64(self):
        tau = 2 * pi
        expected = b"\xC1" + struct.pack(">d", tau)
        self.assert_packable(tau, expected)

    def test_positive_float64(self):
        for e in range(0, 100):
            r = float(2 ** e) + 0.5
            expected = b"\xC1" + struct.pack(">d", r)
            self.assert_packable(r, expected)

    def test_negative_float64(self):
        for e in range(0, 100):
            r = -(float(2 ** e) + 0.5)
            expected = b"\xC1" + struct.pack(">d", r)
            self.assert_packable(r, expected)

    def test_empty_bytes(self):
        self.assert_packable(b"", b"\xCC\x00")

    def test_empty_bytearray(self):
        self.assert_packable(bytearray(), b"\xCC\x00")

    def test_bytes_8(self):
        self.assert_packable(bytearray(b"hello"), b"\xCC\x05hello")

    def test_bytes_16(self):
        b = bytearray(40000)
        self.assert_packable(b, b"\xCD\x9C\x40" + b)

    def test_bytes_32(self):
        b = bytearray(80000)
        self.assert_packable(b, b"\xCE\x00\x01\x38\x80" + b)

    def test_bytearray_size_overflow(self):
        stream_out = BytesIO()
        packer = Packer(stream_out)
        with raises(OverflowError):
            packer.pack_bytes_header(2 ** 32)

    def test_empty_string(self):
        self.assert_packable(u"", b"\x80")

    def test_tiny_strings(self):
        for size in range(0x10):
            self.assert_packable(u"A" * size, bytes(bytearray([0x80 + size]) + (b"A" * size)))

    def test_string_8(self):
        t = u"A" * 40
        b = t.encode("utf-8")
        self.assert_packable(t, b"\xD0\x28" + b)

    def test_string_16(self):
        t = u"A" * 40000
        b = t.encode("utf-8")
        self.assert_packable(t, b"\xD1\x9C\x40" + b)

    def test_string_32(self):
        t = u"A" * 80000
        b = t.encode("utf-8")
        self.assert_packable(t, b"\xD2\x00\x01\x38\x80" + b)

    def test_unicode_string(self):
        t = u"héllö"
        b = t.encode("utf-8")
        self.assert_packable(t, bytes(bytearray([0x80 + len(b)])) + b)

    def test_string_size_overflow(self):
        stream_out = BytesIO()
        packer = Packer(stream_out)
        with raises(OverflowError):
            packer.pack_string_header(2 ** 32)

    def test_empty_list(self):
        self.assert_packable([], b"\x90")

    def test_tiny_lists(self):
        for size in range(0x10):
            data_out = bytearray([0x90 + size]) + bytearray([1] * size)
            self.assert_packable([1] * size, bytes(data_out))

    def test_list_8(self):
        l = [1] * 40
        self.assert_packable(l, b"\xD4\x28" + (b"\x01" * 40))

    def test_list_16(self):
        l = [1] * 40000
        self.assert_packable(l, b"\xD5\x9C\x40" + (b"\x01" * 40000))

    def test_list_32(self):
        l = [1] * 80000
        self.assert_packable(l, b"\xD6\x00\x01\x38\x80" + (b"\x01" * 80000))

    def test_nested_lists(self):
        self.assert_packable([[[]]], b"\x91\x91\x90")

    def test_list_stream(self):
        packed_value = b"\xD7\x01\x02\x03\xDF"
        unpacked_value = [1, 2, 3]
        stream_out = BytesIO()
        packer = Packer(stream_out)
        packer.pack_list_stream_header()
        packer.pack(1)
        packer.pack(2)
        packer.pack(3)
        packer.pack_end_of_stream()
        packed = stream_out.getvalue()
        try:
            assert packed == packed_value
        except AssertionError:
            raise AssertionError("Packed value is %r instead of expected %r" %
                                 (packed, packed_value))
        unpacked = Unpacker(UnpackableBuffer(packed)).unpack()
        try:
            assert unpacked == unpacked_value
        except AssertionError:
            raise AssertionError("Unpacked value %r is not equal to expected %r" %
                                 (unpacked, unpacked_value))

    def test_list_size_overflow(self):
        stream_out = BytesIO()
        packer = Packer(stream_out)
        with raises(OverflowError):
            packer.pack_list_header(2 ** 32)

    def test_empty_map(self):
        self.assert_packable({}, b"\xA0")

    def test_tiny_maps(self):
        for size in range(0x10):
            data_in = OrderedDict()
            data_out = bytearray([0xA0 + size])
            for el in range(1, size + 1):
                data_in[chr(64 + el)] = el
                data_out += bytearray([0x81, 64 + el, el])
            self.assert_packable(data_in, bytes(data_out))

    def test_map_8(self):
        d = OrderedDict([(u"A%s" % i, 1) for i in range(40)])
        b = b"".join(self.packb(u"A%s" % i, 1) for i in range(40))
        self.assert_packable(d, b"\xD8\x28" + b)

    def test_map_16(self):
        d = OrderedDict([(u"A%s" % i, 1) for i in range(40000)])
        b = b"".join(self.packb(u"A%s" % i, 1) for i in range(40000))
        self.assert_packable(d, b"\xD9\x9C\x40" + b)

    def test_map_32(self):
        d = OrderedDict([(u"A%s" % i, 1) for i in range(80000)])
        b = b"".join(self.packb(u"A%s" % i, 1) for i in range(80000))
        self.assert_packable(d, b"\xDA\x00\x01\x38\x80" + b)

    def test_map_stream(self):
        packed_value = b"\xDB\x81A\x01\x81B\x02\xDF"
        unpacked_value = {u"A": 1, u"B": 2}
        stream_out = BytesIO()
        packer = Packer(stream_out)
        packer.pack_map_stream_header()
        packer.pack(u"A")
        packer.pack(1)
        packer.pack(u"B")
        packer.pack(2)
        packer.pack_end_of_stream()
        packed = stream_out.getvalue()
        try:
            assert packed == packed_value
        except AssertionError:
            raise AssertionError("Packed value is %r instead of expected %r" %
                                 (packed, packed_value))
        unpacked = Unpacker(UnpackableBuffer(packed)).unpack()
        try:
            assert unpacked == unpacked_value
        except AssertionError:
            raise AssertionError("Unpacked value %r is not equal to expected %r" %
                                 (unpacked, unpacked_value))

    def test_map_size_overflow(self):
        stream_out = BytesIO()
        packer = Packer(stream_out)
        with raises(OverflowError):
            packer.pack_map_header(2 ** 32)

    def test_illegal_signature(self):
        with self.assertRaises(ValueError):
            self.assert_packable(Structure(b"XXX"), b"\xB0XXX")

    def test_empty_struct(self):
        self.assert_packable(Structure(b"X"), b"\xB0X")

    def test_tiny_structs(self):
        for size in range(0x10):
            fields = [1] * size
            data_in = Structure(b"A", *fields)
            data_out = bytearray([0xB0 + size, 0x41] + fields)
            self.assert_packable(data_in, bytes(data_out))

    def test_struct_size_overflow(self):
        with raises(OverflowError):
            fields = [1] * 16
            self.packb(Structure(b"X", *fields))

    def test_illegal_uuid(self):
        with self.assertRaises(ValueError):
            self.assert_packable(uuid4(), b"\xB0XXX")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from contextlib import redirect_stdout
from io import StringIO

# tag::hello-world-import[]
from neo4j import GraphDatabase
# end::hello-world-import[]

from neo4j.exceptions import ServiceUnavailable
from neo4j._exceptions import BoltHandshakeError


# python -m pytest tests/integration/examples/test_hello_world_example.py -s -v

# tag::hello-world[]
class HelloWorldExample:

    def __init__(self, uri, user, password):
        self.driver = GraphDatabase.driver(uri, auth=(user, password))

    def close(self):
        self.driver.close()

    def print_greeting(self, message):
        with self.driver.session() as session:
            greeting = session.write_transaction(self._create_and_return_greeting, message)
            print(greeting)

    @staticmethod
    def _create_and_return_greeting(tx, message):
        result = tx.run("CREATE (a:Greeting) "
                        "SET a.message = $message "
                        "RETURN a.message + ', from node ' + id(a)", message=message)
        return result.single()[0]


if __name__ == "__main__":
    greeter = HelloWorldExample("bolt://localhost:7687", "neo4j", "password")
    greeter.print_greeting("hello, world")
    greeter.close()
# end::hello-world[]

# tag::hello-world-output[]
# hello, world, from node 1234
# end::hello-world-output[]


def test_hello_world_example(uri, auth):
    try:
        s = StringIO()
        with redirect_stdout(s):
            example = HelloWorldExample(uri, auth[0], auth[1])
            example.print_greeting("hello, world")
            example.close()

        assert s.getvalue().startswith("hello, world, from node ")
    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])




#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

# tag::driver-lifecycle-import[]
from neo4j import GraphDatabase
# end::driver-lifecycle-import[]

from neo4j.exceptions import ServiceUnavailable
from neo4j._exceptions import BoltHandshakeError


# python -m pytest tests/integration/examples/test_driver_lifecycle_example.py -s -v

# tag::driver-lifecycle[]
class DriverLifecycleExample:
    def __init__(self, uri, auth):
        self.driver = GraphDatabase.driver(uri, auth=auth)

    def close(self):
        self.driver.close()
# end::driver-lifecycle[]


def test_example(uri, auth):
    try:
        eg = DriverLifecycleExample(uri, auth)
        eg.close()
    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j.graph import (
    Node,
    Relationship,
    Graph,
    Path,
)
from neo4j.exceptions import Neo4jError


def test_result_graph_instance(session):
    # python -m pytest tests/integration/test_result_graph.py -s -v -k test_result_graph_instance
    result = session.run("RETURN 1")
    graph = result.graph()

    assert isinstance(graph, Graph)


def test_result_graph_case_1(session):
    # python -m pytest tests/integration/test_result_graph.py -s -v -k test_result_graph_case_1
    result = session.run("CREATE (n1:Person:LabelTest1 {name:'Alice'})-[r1:KNOWS {since:1999}]->(n2:Person:LabelTest2 {name:'Bob'}) RETURN n1, r1, n2")
    graph = result.graph()
    assert isinstance(graph, Graph)

    node_view = graph.nodes
    relationships_view = graph.relationships

    for node in node_view:
        name = node["name"]
        if name == "Alice":
            assert node.labels == frozenset(["Person", "LabelTest1"])
        elif name == "Bob":
            assert node.labels == frozenset(["Person", "LabelTest2"])
        else:
            pytest.fail("should only contain 2 nodes, Alice and Bob. {}".format(name))

    for relationship in relationships_view:
        since = relationship["since"]
        assert since == 1999
        assert relationship.type == "KNOWS"



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j.exceptions import (
    ConfigurationError,
)
from neo4j.conf import (
    Config,
    PoolConfig,
    WorkspaceConfig,
    SessionConfig,
)
from neo4j.api import (
    TRUST_SYSTEM_CA_SIGNED_CERTIFICATES,
    WRITE_ACCESS,
    READ_ACCESS,
)

# python -m pytest tests/unit/test_conf.py -s -v

from neo4j.debug import watch
watch("neo4j")

test_pool_config = {
    "connection_timeout": 30.0,
    "init_size": 1,
    "keep_alive": True,
    "max_connection_lifetime": 3600,
    "max_connection_pool_size": 100,
    "protocol_version": None,
    "resolver": None,
    "encrypted": False,
    "user_agent": "test",
    "trust": TRUST_SYSTEM_CA_SIGNED_CERTIFICATES,
}

test_session_config = {
    "connection_acquisition_timeout": 60.0,
    "max_transaction_retry_time": 30.0,
    "initial_retry_delay": 1.0,
    "retry_delay_multiplier": 2.0,
    "retry_delay_jitter_factor": 0.2,
    "bookmarks": (),
    "default_access_mode": WRITE_ACCESS,
    "database": None,
    "fetch_size": 100,
}

config_function_names = ["consume_chain", "consume"]


def test_pool_config_consume():

    test_config = dict(test_pool_config)

    consumed_pool_config = PoolConfig.consume(test_config)

    assert isinstance(consumed_pool_config, PoolConfig)

    assert len(test_config) == 0

    for key in test_pool_config.keys():
        assert consumed_pool_config[key] == test_pool_config[key]

    for key in consumed_pool_config.keys():
        if key not in config_function_names:
            assert test_pool_config[key] == consumed_pool_config[key]

    assert len(consumed_pool_config) - len(config_function_names) == len(test_pool_config)


def test_pool_config_consume_default_values():

    test_config = {}

    consumed_pool_config = PoolConfig.consume(test_config)

    assert isinstance(consumed_pool_config, PoolConfig)

    assert len(test_config) == 0

    consumed_pool_config.keep_alive = "changed"

    assert PoolConfig.keep_alive != consumed_pool_config.keep_alive


def test_pool_config_consume_key_not_valid():

    test_config = dict(test_pool_config)

    test_config["not_valid_key"] = "test"

    with pytest.raises(ConfigurationError) as error:
        consumed_pool_config = PoolConfig.consume(test_config)

    error.match("Unexpected config keys: not_valid_key")


def test_pool_config_set_value():

    test_config = dict(test_pool_config)

    consumed_pool_config = PoolConfig.consume(test_config)

    assert consumed_pool_config.get("encrypted") is False
    assert consumed_pool_config["encrypted"] is False
    assert consumed_pool_config.encrypted is False

    consumed_pool_config.encrypted = "test"

    assert consumed_pool_config.get("encrypted") == "test"
    assert consumed_pool_config["encrypted"] == "test"
    assert consumed_pool_config.encrypted == "test"

    consumed_pool_config.not_valid_key = "test"  # Use consume functions


def test_pool_config_consume_and_then_consume_again():

    test_config = dict(test_pool_config)
    consumed_pool_config = PoolConfig.consume(test_config)
    assert consumed_pool_config.encrypted is False
    consumed_pool_config.encrypted = "test"

    with pytest.raises(AttributeError):
        consumed_pool_config = PoolConfig.consume(consumed_pool_config)

    consumed_pool_config = PoolConfig.consume(dict(consumed_pool_config.items()))
    consumed_pool_config = PoolConfig.consume(dict(consumed_pool_config.items()))

    assert consumed_pool_config.encrypted == "test"


def test_config_consume_chain():

    test_config = {}

    test_config.update(test_pool_config)

    test_config.update(test_session_config)

    consumed_pool_config, consumed_session_config = Config.consume_chain(test_config, PoolConfig, SessionConfig)

    assert isinstance(consumed_pool_config, PoolConfig)
    assert isinstance(consumed_session_config, SessionConfig)

    assert len(test_config) == 0

    for key, val in test_pool_config.items():
        assert consumed_pool_config[key] == val

    for key, val in consumed_pool_config.items():
        if key not in config_function_names:
            assert test_pool_config[key] == val

    assert len(consumed_pool_config) - len(config_function_names) == len(test_pool_config)

    assert len(consumed_session_config) - len(config_function_names) == len(test_session_config)


def test_init_session_config_merge():
    # python -m pytest tests/unit/test_conf.py -s -v -k test_init_session_config

    test_config_a = {"connection_acquisition_timeout": 111}
    test_config_c = {"max_transaction_retry_time": 222}

    workspace_config = WorkspaceConfig(test_config_a, WorkspaceConfig.consume(test_config_c))
    assert len(test_config_a) == 1
    assert len(test_config_c) == 0
    assert isinstance(workspace_config, WorkspaceConfig)
    assert workspace_config.connection_acquisition_timeout == WorkspaceConfig.connection_acquisition_timeout
    assert workspace_config.max_transaction_retry_time == 222

    workspace_config = WorkspaceConfig(test_config_c, test_config_a)
    assert isinstance(workspace_config, WorkspaceConfig)
    assert workspace_config.connection_acquisition_timeout == 111
    assert workspace_config.max_transaction_retry_time == WorkspaceConfig.max_transaction_retry_time

    test_config_b = {"default_access_mode": READ_ACCESS, "connection_acquisition_timeout": 333}

    session_config = SessionConfig(workspace_config, test_config_b)
    assert session_config.connection_acquisition_timeout == 333
    assert session_config.default_access_mode == READ_ACCESS

    session_config = SessionConfig(test_config_b, workspace_config)
    assert session_config.connection_acquisition_timeout == 111
    assert session_config.default_access_mode == READ_ACCESS


def test_init_session_config_with_not_valid_key():
    # python -m pytest tests/unit/test_conf.py -s -v -k test_init_session_config_with_not_valid_key

    test_config_a = {"connection_acquisition_timeout": 111}
    workspace_config = WorkspaceConfig.consume(test_config_a)

    test_config_b = {"default_access_mode": READ_ACCESS, "connection_acquisition_timeout": 333, "not_valid_key": None}
    session_config = SessionConfig(workspace_config, test_config_b)

    with pytest.raises(AttributeError):
        assert session_config.not_valid_key is None

    with pytest.raises(ConfigurationError):
        _ = SessionConfig.consume(test_config_b)

    assert session_config.connection_acquisition_timeout == 333



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import deque
from logging import getLogger
from random import random
from time import perf_counter, sleep
from warnings import warn

from neo4j.conf import SessionConfig
from neo4j.api import READ_ACCESS, WRITE_ACCESS
from neo4j.data import DataHydrator, DataDehydrator
from neo4j.exceptions import (
    Neo4jError,
    ServiceUnavailable,
    TransientError,
    SessionExpired,
    TransactionError,
    ClientError,
)
from neo4j._exceptions import BoltIncompleteCommitError
from neo4j.work import Workspace
from neo4j.work.summary import ResultSummary
from neo4j.work.transaction import Transaction
from neo4j.work.result import Result
from neo4j.io._bolt3 import Bolt3

log = getLogger("neo4j")


class Session(Workspace):
    """A :class:`.Session` is a logical context for transactional units
    of work. Connections are drawn from the :class:`.Driver` connection
    pool as required.

    Session creation is a lightweight operation and sessions are not thread
    safe. Therefore a session should generally be short-lived, and not
    span multiple threads.

    In general, sessions will be created and destroyed within a `with`
    context. For example::

        with driver.session() as session:
            result = session.run("MATCH (n:Person) RETURN n.name AS name")
            # do something with the result...

    :param pool: connection pool instance
    :param config: session config instance
    """

    # The current connection.
    _connection = None

    # The current :class:`.Transaction` instance, if any.
    _transaction = None

    # The current auto-transaction result, if any.
    _autoResult = None

    # The set of bookmarks after which the next
    # :class:`.Transaction` should be carried out.
    _bookmarks = None

    # The state this session is in.
    _state_failed = False

    # Session have been properly closed.
    _closed = False

    def __init__(self, pool, session_config):
        super().__init__(pool, session_config)
        assert isinstance(session_config, SessionConfig)
        self._bookmarks = tuple(session_config.bookmarks)

    def __del__(self):
        try:
            self.close()
        except OSError:
            pass

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        if exception_type:
            self._state_failed = True
        self.close()

    def _connect(self, access_mode, database):
        if access_mode is None:
            access_mode = self._config.default_access_mode
        if self._connection:
            # TODO: Investigate this
            # log.warning("FIXME: should always disconnect before connect")
            self._connection.send_all()
            self._connection.fetch_all()
            self._disconnect()
        self._connection = self._pool.acquire(access_mode=access_mode, timeout=self._config.connection_acquisition_timeout, database=database)

    def _disconnect(self):
        if self._connection:
            self._connection.in_use = False
            self._connection = None

    def _collect_bookmark(self, bookmark):
        if bookmark:
            self._bookmarks = [bookmark]

    def _result_closed(self):
        if self._autoResult:
            self._collect_bookmark(self._autoResult._bookmark)
            self._autoResult = None
            self._disconnect()

    def close(self):
        """Close the session. This will release any borrowed resources, such as connections, and will roll back any outstanding transactions.
        """
        if self._connection:
            if self._autoResult:
                if self._state_failed is False:
                    try:
                        self._autoResult.consume()
                        self._collect_bookmark(self._autoResult._bookmark)
                    except Exception as error:
                        # TODO: Investigate potential non graceful close states
                        self._autoResult = None
                        self._state_failed = True

            if self._transaction:
                if self._transaction.closed() is False:
                    self._transaction.rollback()  # roll back the transaction if it is not closed
                self._transaction = None

            try:
                if self._connection:
                    self._connection.send_all()
                    self._connection.fetch_all()
                    # TODO: Investigate potential non graceful close states
            except Neo4jError:
                pass
            except TransactionError:
                pass
            except ServiceUnavailable:
                pass
            except SessionExpired:
                pass
            finally:
                self._disconnect()

            self._state_failed = False
            self._closed = True

    def run(self, query, parameters=None, **kwparameters):
        """Run a Cypher query within an auto-commit transaction.

        The query is sent and the result header received
        immediately but the :class:`neo4j.Result` content is
        fetched lazily as consumed by the client application.

        If a query is executed before a previous
        :class:`neo4j.Result` in the same :class:`.Session` has
        been fully consumed, the first result will be fully fetched
        and buffered. Note therefore that the generally recommended
        pattern of usage is to fully consume one result before
        executing a subsequent query. If two results need to be
        consumed in parallel, multiple :class:`.Session` objects
        can be used as an alternative to result buffering.

        For more usage details, see :meth:`.Transaction.run`.

        :param query: cypher query
        :type query: str, neo4j.Query
        :param parameters: dictionary of parameters
        :type parameters: dict
        :param kwparameters: additional keyword parameters
        :returns: a new :class:`neo4j.Result` object
        :rtype: :class:`neo4j.Result`
        """
        if not query:
            raise ValueError("Cannot run an empty query")
        if not isinstance(query, (str, Query)):
            raise TypeError("query must be a string or a Query instance")

        if self._transaction:
            raise ClientError("Explicit Transaction must be handled explicitly")

        if self._autoResult:
            self._autoResult._buffer_all()  # This will buffer upp all records for the previous auto-transaction

        if not self._connection:
            self._connect(self._config.default_access_mode, database=self._config.database)
        cx = self._connection
        protocol_version = cx.PROTOCOL_VERSION
        server_info = cx.server_info

        hydrant = DataHydrator()

        self._autoResult = Result(cx, hydrant, self._config.fetch_size, self._result_closed)
        self._autoResult._run(query, parameters, self._config.database, self._config.default_access_mode, self._bookmarks, **kwparameters)

        return self._autoResult

    def last_bookmark(self):
        """Return the bookmark received following the last completed transaction.
        Note: For auto-transaction (Session.run) this will trigger an consume for the current result.

        :returns: :class:`neo4j.Bookmark` object
        """
        # The set of bookmarks to be passed into the next transaction.

        if self._autoResult:
            self._autoResult.consume()

        if self._transaction and self._transaction._closed:
            self._collect_bookmark(self._transaction._bookmark)
            self._transaction = None

        if len(self._bookmarks):
            return self._bookmarks[len(self._bookmarks)-1]
        return None

    def _transaction_closed_handler(self):
        if self._transaction:
            self._collect_bookmark(self._transaction._bookmark)
            self._transaction = None
            self._disconnect()

    def _open_transaction(self, *, access_mode, database, metadata=None, timeout=None):
        self._connect(access_mode=access_mode, database=database)
        self._transaction = Transaction(self._connection, self._config.fetch_size, self._transaction_closed_handler)
        self._transaction._begin(database, self._bookmarks, access_mode, metadata, timeout)

    def begin_transaction(self, metadata=None, timeout=None):
        """ Begin a new unmanaged transaction. Creates a new :class:`.Transaction` within this session.
            At most one transaction may exist in a session at any point in time.
            To maintain multiple concurrent transactions, use multiple concurrent sessions.

            Note: For auto-transaction (Session.run) this will trigger an consume for the current result.

        :param metadata:
            a dictionary with metadata.
            Specified metadata will be attached to the executing transaction and visible in the output of ``dbms.listQueries`` and ``dbms.listTransactions`` procedures.
            It will also get logged to the ``query.log``.
            This functionality makes it easier to tag transactions and is equivalent to ``dbms.setTXMetaData`` procedure, see https://neo4j.com/docs/operations-manual/current/reference/procedures/ for procedure reference.
        :type metadata: dict

        :param timeout:
            the transaction timeout in milliseconds.
            Transactions that execute longer than the configured timeout will be terminated by the database.
            This functionality allows to limit query/transaction execution time.
            Specified timeout overrides the default timeout configured in the database using ``dbms.transaction.timeout`` setting.
            Value should not represent a duration of zero or negative duration.
        :type timeout: int

        :returns: A new transaction instance.
        :rtype: :class:`neo4j.Transaction`

        :raises TransactionError: :class:`neo4j.exceptions.TransactionError` if a transaction is already open.
        """
        # TODO: Implement TransactionConfig consumption

        if self._autoResult:
            self._autoResult.consume()

        if self._transaction:
            raise TransactionError("Explicit transaction already open")

        self._open_transaction(access_mode=self._config.default_access_mode, database=self._config.database, metadata=metadata, timeout=timeout)

        return self._transaction

    def _run_transaction(self, access_mode, transaction_function, *args, **kwargs):

        if not callable(transaction_function):
            raise TypeError("Unit of work is not callable")

        metadata = getattr(transaction_function, "metadata", None)
        timeout = getattr(transaction_function, "timeout", None)

        retry_delay = retry_delay_generator(self._config.initial_retry_delay, self._config.retry_delay_multiplier, self._config.retry_delay_jitter_factor)

        errors = []

        t0 = -1  # Timer

        while True:
            try:
                self._open_transaction(access_mode=access_mode, database=self._config.database, metadata=metadata, timeout=timeout)
                tx = self._transaction
                try:
                    result = transaction_function(tx, *args, **kwargs)
                except Exception:
                    tx.rollback()
                    raise
                else:
                    tx.commit()
            except (ServiceUnavailable, SessionExpired) as error:
                errors.append(error)
                self._disconnect()
            except TransientError as transient_error:
                if not transient_error.is_retriable():
                    raise
                errors.append(transient_error)
            else:
                return result
            if t0 == -1:
                t0 = perf_counter()  # The timer should be started after the first attempt
            t1 = perf_counter()
            if t1 - t0 > self._config.max_transaction_retry_time:
                break
            delay = next(retry_delay)
            log.warning("Transaction failed and will be retried in {}s ({})".format(delay, "; ".join(errors[-1].args)))
            sleep(delay)

        if errors:
            raise errors[-1]
        else:
            raise ServiceUnavailable("Transaction failed")

    def read_transaction(self, transaction_function, *args, **kwargs):
        """Execute a unit of work in a managed read transaction.
        This transaction will automatically be committed unless an exception is thrown during query execution or by the user code.
        Note, that this function perform retries and that the supplied `transaction_function` might get invoked more than once.

        Managed transactions should not generally be explicitly committed (via tx.commit()).

        Example::

            def do_cypher_tx(tx, cypher):
                result = tx.run(cypher)
                values = []
                for record in result:
                    values.append(record.values())
                return values

            with driver.session() as session:
                values = session.read_transaction(do_cypher_tx, "RETURN 1 AS x")

        Example::

            def get_two_tx(tx):
                result = tx.run("UNWIND [1,2,3,4] AS x RETURN x")
                values = []
                for ix, record in enumerate(result):
                    if x > 1:
                        break
                    values.append(record.values())
                info = result.consume()  # discard the remaining records if there are any
                # use the info for logging etc.
                return values

            with driver.session() as session:
                values = session.read_transaction(get_two_tx)

        :param transaction_function: a function that takes a transaction as an argument and does work with the transaction. `tx_function(tx, \*args, \*\*kwargs)`
        :param args: arguments for the `transaction_function`
        :param kwargs: key word arguments for the `transaction_function`
        :return: a result as returned by the given unit of work
        """
        return self._run_transaction(READ_ACCESS, transaction_function, *args, **kwargs)

    def write_transaction(self, transaction_function, *args, **kwargs):
        """Execute a unit of work in a managed write transaction.
        This transaction will automatically be committed unless an exception is thrown during query execution or by the user code.
        Note, that this function perform retries and that the supplied `transaction_function` might get invoked more than once.

        Managed transactions should not generally be explicitly committed (via tx.commit()).

        Example::

            def create_node_tx(tx, name):
                result = tx.run("CREATE (n:NodeExample { name: $name }) RETURN id(n) AS node_id", name=name)
                record = result.single()
                return record["node_id"]

            with driver.session() as session:
                node_id = session.write_transaction(create_node_tx, "example")


        :param transaction_function: a function that takes a transaction as an argument and does work with the transaction. `tx_function(tx, \*args, \*\*kwargs)`
        :param args: key word arguments for the `transaction_function`
        :param kwargs: key word arguments for the `transaction_function`
        :return: a result as returned by the given unit of work
        """
        return self._run_transaction(WRITE_ACCESS, transaction_function, *args, **kwargs)


class Query:
    """ Create a new query.

    :param text: The query text.
    :type text: str
    :param metadata: metadata attached to the query.
    :type metadata: dict
    :param timeout: seconds.
    :type timeout: int
    """
    def __init__(self, text, metadata=None, timeout=None):
        self.text = text

        self.metadata = metadata
        self.timeout = timeout

    def __str__(self):
        return str(self.text)


def unit_of_work(metadata=None, timeout=None):
    """This function is a decorator for transaction functions that allows extra control over how the transaction is carried out.

    For example, a timeout may be applied::

        @unit_of_work(timeout=100)
        def count_people_tx(tx):
            result = tx.run("MATCH (a:Person) RETURN count(a) AS persons")
            record = result.single()
            return record["persons"]

    :param metadata:
        a dictionary with metadata.
        Specified metadata will be attached to the executing transaction and visible in the output of ``dbms.listQueries`` and ``dbms.listTransactions`` procedures.
        It will also get logged to the ``query.log``.
        This functionality makes it easier to tag transactions and is equivalent to ``dbms.setTXMetaData`` procedure, see https://neo4j.com/docs/operations-manual/current/reference/procedures/ for procedure reference.
    :type metadata: dict

    :param timeout:
        the transaction timeout in milliseconds.
        Transactions that execute longer than the configured timeout will be terminated by the database.
        This functionality allows to limit query/transaction execution time.
        Specified timeout overrides the default timeout configured in the database using ``dbms.transaction.timeout`` setting.
        Value should not represent a duration of zero or negative duration.
    :type timeout: int
    """

    def wrapper(f):

        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)

        wrapped.metadata = metadata
        wrapped.timeout = timeout
        return wrapped

    return wrapper


def retry_delay_generator(initial_delay, multiplier, jitter_factor):
    delay = initial_delay
    while True:
        jitter = jitter_factor * delay
        yield delay - jitter + (2 * jitter * random())
        delay *= multiplier



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest


from neo4j.spatial import (
    CartesianPoint,
    WGS84Point,
)


def test_cartesian_point_input(cypher_eval):
    x, y = cypher_eval("CYPHER runtime=interpreted "
                       "WITH $point AS point "
                       "RETURN [point.x, point.y]",
                       point=CartesianPoint((1.23, 4.56)))
    assert x == 1.23
    assert y == 4.56


def test_cartesian_3d_point_input(cypher_eval):
    x, y, z = cypher_eval("CYPHER runtime=interpreted "
                          "WITH $point AS point "
                          "RETURN [point.x, point.y, point.z]",
                          point=CartesianPoint((1.23, 4.56, 7.89)))
    assert x == 1.23
    assert y == 4.56
    assert z == 7.89


def test_wgs84_point_input(cypher_eval):
    lat, long = cypher_eval("CYPHER runtime=interpreted "
                            "WITH $point AS point "
                            "RETURN [point.latitude, point.longitude]",
                            point=WGS84Point((1.23, 4.56)))
    assert long == 1.23
    assert lat == 4.56


def test_wgs84_3d_point_input(cypher_eval):
    lat, long, height = cypher_eval("CYPHER runtime=interpreted "
                                    "WITH $point AS point "
                                    "RETURN [point.latitude, point.longitude, "
                                    "point.height]",
                                    point=WGS84Point((1.23, 4.56, 7.89)))
    assert long == 1.23
    assert lat == 4.56
    assert height == 7.89


def test_point_array_input(cypher_eval):
    data = [WGS84Point((1.23, 4.56)), WGS84Point((9.87, 6.54))]
    value = cypher_eval("CREATE (a {x:$x}) RETURN a.x", x=data)
    assert value == data


def test_cartesian_point_output(cypher_eval):
    value = cypher_eval("RETURN point({x:3, y:4})")
    assert isinstance(value, CartesianPoint)
    assert value.x == 3.0
    assert value.y == 4.0
    with pytest.raises(AttributeError):
        _ = value.z


def test_cartesian_3d_point_output(cypher_eval):
    value = cypher_eval("RETURN point({x:3, y:4, z:5})")
    assert isinstance(value, CartesianPoint)
    assert value.x == 3.0
    assert value.y == 4.0
    assert value.z == 5.0


def test_wgs84_point_output(cypher_eval):
    value = cypher_eval("RETURN point({latitude:3, longitude:4})")
    assert isinstance(value, WGS84Point)
    assert value.latitude == 3.0
    assert value.y == 3.0
    assert value.longitude == 4.0
    assert value.x == 4.0
    with pytest.raises(AttributeError):
        _ = value.height
    with pytest.raises(AttributeError):
        _ = value.z


def test_wgs84_3d_point_output(cypher_eval):
    value = cypher_eval("RETURN point({latitude:3, longitude:4, height:5})")
    assert isinstance(value, WGS84Point)
    assert value.latitude == 3.0
    assert value.y == 3.0
    assert value.longitude == 4.0
    assert value.x == 4.0
    assert value.height == 5.0
    assert value.z == 5.0



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
from collections.abc import MutableSet
from logging import getLogger
from time import perf_counter

from neo4j.addressing import Address


log = getLogger("neo4j")


class OrderedSet(MutableSet):

    def __init__(self, elements=()):
        self._elements = OrderedDict.fromkeys(elements)
        self._current = None

    def __repr__(self):
        return "{%s}" % ", ".join(map(repr, self._elements))

    def __contains__(self, element):
        return element in self._elements

    def __iter__(self):
        return iter(self._elements)

    def __len__(self):
        return len(self._elements)

    def __getitem__(self, index):
        return list(self._elements.keys())[index]

    def add(self, element):
        self._elements[element] = None

    def clear(self):
        self._elements.clear()

    def discard(self, element):
        try:
            del self._elements[element]
        except KeyError:
            pass

    def remove(self, element):
        try:
            del self._elements[element]
        except KeyError:
            raise ValueError(element)

    def update(self, elements=()):
        self._elements.update(OrderedDict.fromkeys(elements))

    def replace(self, elements=()):
        e = self._elements
        e.clear()
        e.update(OrderedDict.fromkeys(elements))


class RoutingTable:

    @classmethod
    def parse_routing_info(cls, *, database, servers, ttl):
        """ Parse the records returned from the procedure call and
        return a new RoutingTable instance.
        """
        routers = []
        readers = []
        writers = []
        try:
            for server in servers:
                role = server["role"]
                addresses = []
                for address in server["addresses"]:
                    addresses.append(Address.parse(address, default_port=7687))
                if role == "ROUTE":
                    routers.extend(addresses)
                elif role == "READ":
                    readers.extend(addresses)
                elif role == "WRITE":
                    writers.extend(addresses)
        except (KeyError, TypeError):
            raise ValueError("Cannot parse routing info")
        else:
            return cls(database=database, routers=routers, readers=readers, writers=writers, ttl=ttl)

    def __init__(self, *, database, routers=(), readers=(), writers=(), ttl=0):
        self.initial_routers = OrderedSet(routers)
        self.routers = OrderedSet(routers)
        self.readers = OrderedSet(readers)
        self.writers = OrderedSet(writers)
        self.last_updated_time = perf_counter()
        self.ttl = ttl
        self.database = database

    def __repr__(self):
        return "RoutingTable(database=%r routers=%r, readers=%r, writers=%r, last_updated_time=%r, ttl=%r)" % (
            self.database,
            self.routers,
            self.readers,
            self.writers,
            self.last_updated_time,
            self.ttl,
        )

    def __contains__(self, address):
        return address in self.routers or address in self.readers or address in self.writers

    def is_fresh(self, readonly=False):
        """ Indicator for whether routing information is still usable.
        """
        assert isinstance(readonly, bool)
        log.debug("[#0000]  C: <ROUTING> Checking table freshness (readonly=%r)", readonly)
        expired = self.last_updated_time + self.ttl <= perf_counter()
        if readonly:
            has_server_for_mode = bool(self.readers)
        else:
            has_server_for_mode = bool(self.writers)
        log.debug("[#0000]  C: <ROUTING> Table expired=%r", expired)
        log.debug("[#0000]  C: <ROUTING> Table routers=%r", self.routers)
        log.debug("[#0000]  C: <ROUTING> Table has_server_for_mode=%r", has_server_for_mode)
        return not expired and self.routers and has_server_for_mode

    def missing_fresh_writer(self):
        """ Check if the routing table have a fresh write address.

        :return: Return true if it does not have a fresh write address.
        :rtype: bool
        """
        return not self.is_fresh(readonly=False)

    def should_be_purged_from_memory(self):
        """ Check if the routing table is stale and not used for a long time and should be removed from memory.

        :return: Returns true if it is old and not used for a while.
        :rtype: bool
        """
        from neo4j.conf import RoutingConfig
        perf_time = perf_counter()
        log.debug("[#0000]  C: <ROUTING AGED> last_updated_time=%r perf_time=%r", self.last_updated_time, perf_time)
        return self.last_updated_time + self.ttl + RoutingConfig.routing_table_purge_delay <= perf_time

    def update(self, new_routing_table):
        """ Update the current routing table with new routing information
        from a replacement table.
        """
        self.routers.replace(new_routing_table.routers)
        self.readers.replace(new_routing_table.readers)
        self.writers.replace(new_routing_table.writers)
        self.last_updated_time = perf_counter()
        self.ttl = new_routing_table.ttl
        log.debug("[#0000]  S: <ROUTING> table=%r", self)

    def servers(self):
        return set(self.routers) | set(self.writers) | set(self.readers)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
This module contains the low-level functionality required for speaking
Bolt. It is not intended to be used directly by driver users. Instead,
the `session` module provides the main user-facing abstractions.
"""


__all__ = [
    "Bolt",
    "BoltPool",
    "Neo4jPool",
]


from collections import deque
from logging import getLogger
from random import choice
from select import select
from time import perf_counter

from socket import (
    socket,
    SOL_SOCKET,
    SO_KEEPALIVE,
    SHUT_RDWR,
    timeout as SocketTimeout,
    AF_INET,
    AF_INET6,
)

from ssl import (
    HAS_SNI,
    SSLError,
)

from struct import (
    pack as struct_pack,
)

from threading import (
    Lock,
    RLock,
    Condition,
)

from neo4j.addressing import Address
from neo4j.conf import PoolConfig
from neo4j._exceptions import (
    BoltRoutingError,
    BoltSecurityError,
    BoltProtocolError,
    BoltHandshakeError,
)
from neo4j.exceptions import (
    ServiceUnavailable,
    ClientError,
    SessionExpired,
    ReadServiceUnavailable,
    WriteServiceUnavailable,
    ConfigurationError,
)
from neo4j.routing import RoutingTable
from neo4j.conf import (
    PoolConfig,
    WorkspaceConfig,
)
from neo4j.api import (
    READ_ACCESS,
    WRITE_ACCESS,
)

# Set up logger
log = getLogger("neo4j")


class Bolt:
    """ Server connection for Bolt protocol.

    A :class:`.Bolt` should be constructed following a
    successful .open()

    Bolt handshake and takes the socket over which
    the handshake was carried out.
    """

    MAGIC_PREAMBLE = b"\x60\x60\xB0\x17"

    PROTOCOL_VERSION = None

    @classmethod
    def get_handshake(cls):
        """ Return the supported Bolt versions as bytes.
        The length is 16 bytes as specified in the Bolt version negotiation.
        :return: bytes
        """
        offered_versions = sorted(cls.protocol_handlers().keys(), reverse=True)[:4]
        return b"".join(version.to_bytes() for version in offered_versions).ljust(16, b"\x00")

    @classmethod
    def protocol_handlers(cls, protocol_version=None):
        """ Return a dictionary of available Bolt protocol handlers,
        keyed by version tuple. If an explicit protocol version is
        provided, the dictionary will contain either zero or one items,
        depending on whether that version is supported. If no protocol
        version is provided, all available versions will be returned.

        :param protocol_version: tuple identifying a specific protocol
            version (e.g. (3, 5)) or None
        :return: dictionary of version tuple to handler class for all
            relevant and supported protocol versions
        :raise TypeError: if protocol version is not passed in a tuple
        """

        # Carry out Bolt subclass imports locally to avoid circular dependency issues.
        from neo4j.io._bolt3 import Bolt3
        from neo4j.io._bolt4 import Bolt4x0, Bolt4x1, Bolt4x2

        handlers = {
            Bolt3.PROTOCOL_VERSION: Bolt3,
            Bolt4x0.PROTOCOL_VERSION: Bolt4x0,
            Bolt4x1.PROTOCOL_VERSION: Bolt4x1,
            Bolt4x2.PROTOCOL_VERSION: Bolt4x2,
        }

        if protocol_version is None:
            return handlers

        if not isinstance(protocol_version, tuple):
            raise TypeError("Protocol version must be specified as a tuple")

        if protocol_version in handlers:
            return {protocol_version: handlers[protocol_version]}

        return {}

    @classmethod
    def ping(cls, address, *, timeout=None, **config):
        """ Attempt to establish a Bolt connection, returning the
        agreed Bolt protocol version if successful.
        """
        config = PoolConfig.consume(config)
        try:
            s, protocol_version, handshake, data = connect(
                address,
                timeout=timeout,
                custom_resolver=config.resolver,
                ssl_context=config.get_ssl_context(),
                keep_alive=config.keep_alive,
            )
        except ServiceUnavailable:
            return None
        except BoltHandshakeError as e:
            return None
        else:
            s.close()
            return protocol_version

    @classmethod
    def open(cls, address, *, auth=None, timeout=None, routing_context=None, **pool_config):
        """ Open a new Bolt connection to a given server address.

        :param address:
        :param auth:
        :param timeout: the connection timeout in seconds
        :param routing_context: dict containing routing context
        :param pool_config:
        :return:
        :raise BoltHandshakeError: raised if the Bolt Protocol can not negotiate a protocol version.
        :raise ServiceUnavailable: raised if there was a connection issue.
        """
        pool_config = PoolConfig.consume(pool_config)
        s, pool_config.protocol_version, handshake, data = connect(
            address,
            timeout=timeout,
            custom_resolver=pool_config.resolver,
            ssl_context=pool_config.get_ssl_context(),
            keep_alive=pool_config.keep_alive,
        )

        if pool_config.protocol_version == (3, 0):
            # Carry out Bolt subclass imports locally to avoid circular dependency issues.
            from neo4j.io._bolt3 import Bolt3
            connection = Bolt3(address, s, pool_config.max_connection_lifetime, auth=auth, user_agent=pool_config.user_agent, routing_context=routing_context)
        elif pool_config.protocol_version == (4, 0):
            # Carry out Bolt subclass imports locally to avoid circular dependency issues.
            from neo4j.io._bolt4 import Bolt4x0
            connection = Bolt4x0(address, s, pool_config.max_connection_lifetime, auth=auth, user_agent=pool_config.user_agent, routing_context=routing_context)
        elif pool_config.protocol_version == (4, 1):
            # Carry out Bolt subclass imports locally to avoid circular dependency issues.
            from neo4j.io._bolt4 import Bolt4x1
            connection = Bolt4x1(address, s, pool_config.max_connection_lifetime, auth=auth, user_agent=pool_config.user_agent, routing_context=routing_context)
        elif pool_config.protocol_version == (4, 2):
            # Carry out Bolt subclass imports locally to avoid circular dependency issues.
            from neo4j.io._bolt4 import Bolt4x2
            connection = Bolt4x2(address, s, pool_config.max_connection_lifetime, auth=auth, user_agent=pool_config.user_agent, routing_context=routing_context)
        else:
            log.debug("[#%04X]  S: <CLOSE>", s.getpeername()[1])
            s.shutdown(SHUT_RDWR)
            s.close()

            supported_versions = Bolt.protocol_handlers().keys()
            raise BoltHandshakeError("The Neo4J server does not support communication with this driver. This driver have support for Bolt Protocols {}".format(supported_versions), address=address, request_data=handshake, response_data=data)

        try:
            connection.hello()
        except Exception as error:
            log.debug("[#%04X]  C: <CLOSE> %s", s.getsockname()[1], str(error))
            s.shutdown(SHUT_RDWR)
            s.close()
            raise error

        return connection

    @property
    def encrypted(self):
        raise NotImplementedError

    @property
    def der_encoded_server_certificate(self):
        raise NotImplementedError

    @property
    def local_port(self):
        raise NotImplementedError

    def hello(self):
        raise NotImplementedError

    def __del__(self):
        try:
            self.close()
        except OSError:
            pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def run(self, query, parameters=None, mode=None, bookmarks=None, metadata=None,
            timeout=None, db=None, **handlers):
        """ Appends a RUN message to the output stream.

        :param query: Cypher query string
        :param parameters: dictionary of Cypher parameters
        :param mode: access mode for routing - "READ" or "WRITE" (default)
        :param bookmarks: iterable of bookmark values after which this transaction should begin
        :param metadata: custom metadata dictionary to attach to the transaction
        :param timeout: timeout for transaction execution (seconds)
        :param db: name of the database against which to begin the transaction
        :param handlers: handler functions passed into the returned Response object
        :return: Response object
        """

    def discard(self, n=-1, qid=-1, **handlers):
        """ Appends a DISCARD message to the output stream.

        :param n: number of records to discard, default = -1 (ALL)
        :param qid: query ID to discard for, default = -1 (last query)
        :param handlers: handler functions passed into the returned Response object
        :return: Response object
        """

    def pull(self, n=-1, qid=-1, **handlers):
        """ Appends a PULL message to the output stream.

        :param n: number of records to pull, default = -1 (ALL)
        :param qid: query ID to pull for, default = -1 (last query)
        :param handlers: handler functions passed into the returned Response object
        :return: Response object
        """

    def begin(self, mode=None, bookmarks=None, metadata=None, timeout=None, db=None, **handlers):
        """ Appends a BEGIN message to the output stream.

        :param mode: access mode for routing - "READ" or "WRITE" (default)
        :param bookmarks: iterable of bookmark values after which this transaction should begin
        :param metadata: custom metadata dictionary to attach to the transaction
        :param timeout: timeout for transaction execution (seconds)
        :param db: name of the database against which to begin the transaction
        :param handlers: handler functions passed into the returned Response object
        :return: Response object
        """

    def commit(self, **handlers):
        raise NotImplementedError

    def rollback(self, **handlers):
        raise NotImplementedError

    def reset(self):
        """ Add a RESET message to the outgoing queue, send
        it and consume all remaining messages.
        """
        raise NotImplementedError

    def send_all(self):
        """ Send all queued messages to the server.
        """
        raise NotImplementedError

    def fetch_message(self):
        """ Receive at least one message from the server, if available.

        :return: 2-tuple of number of detail messages and number of summary
                 messages fetched
        """
        raise NotImplementedError

    def timedout(self):
        raise NotImplementedError

    def fetch_all(self):
        """ Fetch all outstanding messages.

        :return: 2-tuple of number of detail messages and number of summary
                 messages fetched
        """
        raise NotImplementedError

    def close(self):
        """ Close the connection.
        """
        raise NotImplementedError

    def closed(self):
        raise NotImplementedError

    def defunct(self):
        raise NotImplementedError


class IOPool:
    """ A collection of connections to one or more server addresses.
    """

    def __init__(self, opener, pool_config, workspace_config):
        assert callable(opener)
        assert isinstance(pool_config, PoolConfig)
        assert isinstance(workspace_config, WorkspaceConfig)

        self.opener = opener
        self.pool_config = pool_config
        self.workspace_config = workspace_config
        self.connections = {}
        self.lock = RLock()
        self.cond = Condition(self.lock)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _acquire(self, address, timeout):
        """ Acquire a connection to a given address from the pool.
        The address supplied should always be an IP address, not
        a host name.

        This method is thread safe.
        """
        t0 = perf_counter()
        if timeout is None:
            timeout = self.workspace_config.connection_acquisition_timeout

        with self.lock:
            try:
                connections = self.connections[address]
            except KeyError:
                connections = self.connections[address] = deque()

            def time_remaining():
                t = timeout - (perf_counter() - t0)
                return t if t > 0 else 0

            while True:
                # try to find a free connection in pool
                for connection in list(connections):
                    if connection.closed() or connection.defunct() or connection.timedout():
                        connections.remove(connection)
                        continue
                    if not connection.in_use:
                        connection.in_use = True
                        return connection
                # all connections in pool are in-use
                infinite_pool_size = (self.pool_config.max_connection_pool_size < 0 or self.pool_config.max_connection_pool_size == float("inf"))
                can_create_new_connection = infinite_pool_size or len(connections) < self.pool_config.max_connection_pool_size
                if can_create_new_connection:
                    timeout = min(self.pool_config.connection_timeout, time_remaining())
                    try:
                        connection = self.opener(address, timeout)
                    except ServiceUnavailable:
                        self.remove(address)
                        raise
                    else:
                        connection.pool = self
                        connection.in_use = True
                        connections.append(connection)
                        return connection

                # failed to obtain a connection from pool because the
                # pool is full and no free connection in the pool
                if time_remaining():
                    self.cond.wait(time_remaining())
                    # if timed out, then we throw error. This time
                    # computation is needed, as with python 2.7, we
                    # cannot tell if the condition is notified or
                    # timed out when we come to this line
                    if not time_remaining():
                        raise ClientError("Failed to obtain a connection from pool "
                                          "within {!r}s".format(timeout))
                else:
                    raise ClientError("Failed to obtain a connection from pool "
                                      "within {!r}s".format(timeout))

    def acquire(self, access_mode=None, timeout=None, database=None):
        """ Acquire a connection to a server that can satisfy a set of parameters.

        :param access_mode:
        :param timeout:
        :param database:
        """

    def release(self, *connections):
        """ Release a connection back into the pool.
        This method is thread safe.
        """
        with self.lock:
            for connection in connections:
                connection.in_use = False
            self.cond.notify_all()

    def in_use_connection_count(self, address):
        """ Count the number of connections currently in use to a given
        address.
        """
        try:
            connections = self.connections[address]
        except KeyError:
            return 0
        else:
            return sum(1 if connection.in_use else 0 for connection in connections)

    def deactivate(self, address):
        """ Deactivate an address from the connection pool, if present, closing
        all idle connection to that address
        """
        with self.lock:
            try:
                connections = self.connections[address]
            except KeyError:  # already removed from the connection pool
                return
            for conn in list(connections):
                if not conn.in_use:
                    connections.remove(conn)
                    try:
                        conn.close()
                    except IOError:
                        pass
            if not connections:
                self.remove(address)

    def on_write_failure(self, address):
        raise WriteServiceUnavailable("No write service available for pool {}".format(self))

    def remove(self, address):
        """ Remove an address from the connection pool, if present, closing
        all connections to that address.
        """
        with self.lock:
            for connection in self.connections.pop(address, ()):
                try:
                    connection.close()
                except IOError:
                    pass

    def close(self):
        """ Close all connections and empty the pool.
        This method is thread safe.
        """
        try:
            with self.lock:
                for address in list(self.connections):
                    self.remove(address)
        except TypeError:
            pass


class BoltPool(IOPool):

    @classmethod
    def open(cls, address, *, auth, pool_config, workspace_config, routing_context=None):
        """Create a new BoltPool

        :param address:
        :param auth:
        :param pool_config:
        :param workspace_config:
        :param routing_context:
        :return: BoltPool
        """

        if routing_context is None:
            routing_context = {}
        elif "address" in routing_context:
            raise ConfigurationError("The key 'address' is reserved for routing context.")
        routing_context["address"] = str(address)

        def opener(addr, timeout):
            return Bolt.open(addr, auth=auth, timeout=timeout, routing_context=routing_context, **pool_config)

        pool = cls(opener, pool_config, workspace_config, routing_context, address)
        seeds = [pool.acquire() for _ in range(pool_config.init_size)]
        pool.release(*seeds)
        return pool

    def __init__(self, opener, pool_config, workspace_config, routing_context, address):
        super(BoltPool, self).__init__(opener, pool_config, workspace_config)
        self.address = address
        self.routing_context = routing_context

    def __repr__(self):
        return "<{} address={!r}>".format(self.__class__.__name__, self.address)

    def acquire(self, access_mode=None, timeout=None, database=None):
        # The access_mode and database is not needed for a direct connection, its just there for consistency.
        return self._acquire(self.address, timeout)


class Neo4jPool(IOPool):
    """ Connection pool with routing table.
    """

    @classmethod
    def open(cls, *addresses, auth, pool_config, workspace_config, routing_context=None):
        """Create a new Neo4jPool

        :param addresses: one or more address as positional argument
        :param auth:
        :param pool_config:
        :param workspace_config:
        :param routing_context:
        :return: Neo4jPool
        """

        address = addresses[0]
        if routing_context is None:
            routing_context = {}
        elif "address" in routing_context:
            raise ConfigurationError("The key 'address' is reserved for routing context.")
        routing_context["address"] = str(address)

        def opener(addr, timeout):
            return Bolt.open(addr, auth=auth, timeout=timeout, routing_context=routing_context, **pool_config)

        pool = cls(opener, pool_config, workspace_config, routing_context, address)

        try:
            pool.update_routing_table(database=workspace_config.database)
        except Exception:
            pool.close()
            raise
        else:
            return pool

    def __init__(self, opener, pool_config, workspace_config, routing_context, address):
        """

        :param opener:
        :param pool_config:
        :param workspace_config:
        :param routing_context: Dictionary with routing information
        :param addresses:
        """
        super(Neo4jPool, self).__init__(opener, pool_config, workspace_config)
        # Each database have a routing table, the default database is a special case.
        log.debug("[#0000]  C: <NEO4J POOL> routing address %r", address)
        self.address = address
        self.routing_tables = {workspace_config.database: RoutingTable(database=workspace_config.database, routers=[address])}
        self.routing_context = routing_context
        self.refresh_lock = Lock()

    def __repr__(self):
        """ The representation shows the initial routing addresses.

        :return: The representation
        :rtype: str
        """
        return "<{} addresses={!r}>".format(self.__class__.__name__, self.get_default_database_initial_router_addresses())

    @property
    def first_initial_routing_address(self):
        return self.get_default_database_initial_router_addresses()[0]

    def get_default_database_initial_router_addresses(self):
        """ Get the initial router addresses for the default database.

        :return:
        :rtype: OrderedSet
        """
        return self.get_routing_table_for_default_database().initial_routers

    def get_default_database_router_addresses(self):
        """ Get the router addresses for the default database.

        :return:
        :rtype: OrderedSet
        """
        return self.get_routing_table_for_default_database().routers

    def get_routing_table_for_default_database(self):
        return self.routing_tables[self.workspace_config.database]

    def create_routing_table(self, database):
        if database not in self.routing_tables:
            self.routing_tables[database] = RoutingTable(database=database, routers=self.get_default_database_initial_router_addresses())

    def fetch_routing_info(self, *, address, timeout, database):
        """ Fetch raw routing info from a given router address.

        :param address: router address
        :param timeout: seconds
        :param database: the data base name to get routing table for
        :param address: the address by which the client initially contacted the server as a hint for inclusion in the returned routing table.

        :return: list of routing records or
                 None if no connection could be established
        :raise ServiceUnavailable: if the server does not support routing or
                                   if routing support is broken
        """

        # The name of system database is fixed and named as "system".
        # It cannot be changed for a single instance or a cluster. (We can reliably assume that the system db exists on each instance.)
        #
        # Database name is NOT case sensitive.
        #
        # Each cluster member will host the exact same databases. For example, if a cluster member A has databases "foo" and
        # "system", then all other members in the cluster should also have and only have "foo" and "system".
        # However at a certain time, the cluster members may or may not up-to-date, as a result, cluster members may contain different databases.
        #
        # Maintain a routing table for each database.
        #
        # Default database is named "neo4j", (this can be renamed on the Neo4j server).
        #
        # Any core member in a cluster can provide a routing table for any database inside the cluster.
        # The seed_url can be used to find all databases in the cluster.
        #
        # If the driver failed to refresh routing table with all known routers, then the driver should retry a few times before it raises a ServiceUnavailable.
        #
        # A valid routing table should at least have one router and one reader.
        #
        # To prevent the routing tables from growing infinitely.
        # Stale/Aged routing tables is removed when there is a failure to obtain a routing table.
        # Remove a routing table if it have been aged, timeout = TTL + RoutingConfig.routing_table_purge_delay

        # Carry out Bolt subclass imports locally to avoid circular dependency issues.
        from neo4j.io._bolt3 import Bolt3
        from neo4j.io._bolt4 import Bolt4x0, Bolt4x1, Bolt4x2

        from neo4j.api import (
            SYSTEM_DATABASE,
            DEFAULT_DATABASE,
            READ_ACCESS,
        )

        metadata = {}
        records = []

        def fail(md):
            if md.get("code") == "Neo.ClientError.Procedure.ProcedureNotFound":
                raise BoltRoutingError("Server does not support routing", address)
            else:
                raise BoltRoutingError("Routing support broken on server", address)

        try:
            with self._acquire(address, timeout) as cx:
                log.debug("[#%04X]  C: <ROUTING> query=%r", cx.local_port, self.routing_context or {})

                if database is None:
                    database = self.workspace_config.database

                cx.run_get_routing_table(on_success=metadata.update, on_failure=fail, database=database)
                cx.pull(on_success=metadata.update, on_records=records.extend)
                cx.send_all()
                cx.fetch_all()
                routing_info = [dict(zip(metadata.get("fields", ()), values)) for values in records]
                log.debug("[#%04X]  S: <ROUTING> info=%r", cx.local_port, routing_info)
            return routing_info
        except BoltRoutingError as error:
            raise ServiceUnavailable(*error.args)
        except ServiceUnavailable:
            self.deactivate(address=address)
            return None

    def fetch_routing_table(self, *, address, timeout, database):
        """ Fetch a routing table from a given router address.

        :param address: router address
        :param timeout: seconds
        :param database: the database name
        :type: str

        :return: a new RoutingTable instance or None if the given router is
                 currently unable to provide routing information

        :raise neo4j.exceptions.ServiceUnavailable: if no writers are available
        :raise neo4j._exceptions.BoltProtocolError: if the routing information received is unusable
        """
        new_routing_info = self.fetch_routing_info(address=address, timeout=timeout, database=database)
        if new_routing_info is None:
            return None
        elif not new_routing_info:
            raise BoltRoutingError("Invalid routing table", address)
        else:
            servers = new_routing_info[0]["servers"]
            ttl = new_routing_info[0]["ttl"]
            new_routing_table = RoutingTable.parse_routing_info(database=database, servers=servers, ttl=ttl)

        # Parse routing info and count the number of each type of server
        num_routers = len(new_routing_table.routers)
        num_readers = len(new_routing_table.readers)

        # num_writers = len(new_routing_table.writers)
        # If no writers are available. This likely indicates a temporary state,
        # such as leader switching, so we should not signal an error.

        # No routers
        if num_routers == 0:
            raise BoltRoutingError("No routing servers returned from server", address)

        # No readers
        if num_readers == 0:
            raise BoltRoutingError("No read servers returned from server", address)

        # At least one of each is fine, so return this table
        return new_routing_table

    def update_routing_table_from(self, *routers, database=None):
        """ Try to update routing tables with the given routers.

        :return: True if the routing table is successfully updated,
        otherwise False
        """
        log.debug("Attempting to update routing table from {}".format(", ".join(map(repr, routers))))
        for router in routers:
            new_routing_table = self.fetch_routing_table(address=router, timeout=self.pool_config.connection_timeout, database=database)
            if new_routing_table is not None:
                self.routing_tables[database].update(new_routing_table)
                log.debug("[#0000]  C: <UPDATE ROUTING TABLE> address={!r} ({!r})".format(router, self.routing_tables[database]))
                return True
        return False

    def update_routing_table(self, *, database):
        """ Update the routing table from the first router able to provide
        valid routing information.

        :param database: The database name

        :raise neo4j.exceptions.ServiceUnavailable:
        """
        # copied because it can be modified
        existing_routers = list(self.routing_tables[database].routers)

        has_tried_initial_routers = False
        if self.routing_tables[database].missing_fresh_writer():
            # TODO: Test this state
            has_tried_initial_routers = True
            if self.update_routing_table_from(self.first_initial_routing_address, database=database):
                # Why is only the first initial routing address used?
                return

        if self.update_routing_table_from(*existing_routers, database=database):
            return

        if not has_tried_initial_routers and self.first_initial_routing_address not in existing_routers:
            if self.update_routing_table_from(self.first_initial_routing_address, database=database):
                # Why is only the first initial routing address used?
                return

        # None of the routers have been successful, so just fail
        log.error("Unable to retrieve routing information")
        raise ServiceUnavailable("Unable to retrieve routing information")

    def update_connection_pool(self, *, database):
        servers = self.routing_tables[database].servers()
        for address in list(self.connections):
            if address not in servers:
                super(Neo4jPool, self).deactivate(address)

    def ensure_routing_table_is_fresh(self, *, access_mode, database):
        """ Update the routing table if stale.

        This method performs two freshness checks, before and after acquiring
        the refresh lock. If the routing table is already fresh on entry, the
        method exits immediately; otherwise, the refresh lock is acquired and
        the second freshness check that follows determines whether an update
        is still required.

        This method is thread-safe.

        :return: `True` if an update was required, `False` otherwise.
        """
        from neo4j.api import READ_ACCESS
        if self.routing_tables[database].is_fresh(readonly=(access_mode == READ_ACCESS)):
            # Readers are fresh.
            return False
        with self.refresh_lock:

            self.update_routing_table(database=database)
            self.update_connection_pool(database=database)

            for database in list(self.routing_tables.keys()):
                # Remove unused databases in the routing table
                # Remove the routing table after a timeout = TTL + 30s
                log.debug("[#0000]  C: <ROUTING AGED> database=%s", database)
                if self.routing_tables[database].should_be_purged_from_memory() and database != self.workspace_config.database:
                    del self.routing_tables[database]

            return True

    def _select_address(self, *, access_mode, database):
        from neo4j.api import READ_ACCESS
        """ Selects the address with the fewest in-use connections.
        """
        self.create_routing_table(database)
        self.ensure_routing_table_is_fresh(access_mode=access_mode, database=database)
        log.debug("[#0000]  C: <ROUTING TABLE ENSURE FRESH> %r", self.routing_tables)
        if access_mode == READ_ACCESS:
            addresses = self.routing_tables[database].readers
        else:
            addresses = self.routing_tables[database].writers
        addresses_by_usage = {}
        for address in addresses:
            addresses_by_usage.setdefault(self.in_use_connection_count(address), []).append(address)
        if not addresses_by_usage:
            if access_mode == READ_ACCESS:
                raise ReadServiceUnavailable("No read service currently available")
            else:
                raise WriteServiceUnavailable("No write service currently available")
        return choice(addresses_by_usage[min(addresses_by_usage)])

    def acquire(self, access_mode=None, timeout=None, database=None):
        if access_mode not in (WRITE_ACCESS, READ_ACCESS):
            raise ClientError("Non valid 'access_mode'; {}".format(access_mode))
        if not timeout:
            raise ClientError("'timeout' must be a float larger than 0; {}".format(timeout))

        from neo4j.api import check_access_mode
        access_mode = check_access_mode(access_mode)
        while True:
            try:
                # Get an address for a connection that have the fewest in-use connections.
                address = self._select_address(access_mode=access_mode, database=database)
                log.debug("[#0000]  C: <ACQUIRE ADDRESS> database=%r address=%r", database, address)
            except (ReadServiceUnavailable, WriteServiceUnavailable) as err:
                raise SessionExpired("Failed to obtain connection towards '%s' server." % access_mode) from err
            try:
                connection = self._acquire(address, timeout=timeout)  # should always be a resolved address
            except ServiceUnavailable:
                self.deactivate(address=address)
            else:
                return connection

    def deactivate(self, address):
        """ Deactivate an address from the connection pool,
        if present, remove from the routing table and also closing
        all idle connections to that address.
        """
        log.debug("[#0000]  C: <ROUTING> Deactivating address %r", address)
        # We use `discard` instead of `remove` here since the former
        # will not fail if the address has already been removed.
        for database in self.routing_tables.keys():
            self.routing_tables[database].routers.discard(address)
            self.routing_tables[database].readers.discard(address)
            self.routing_tables[database].writers.discard(address)
        log.debug("[#0000]  C: <ROUTING> table=%r", self.routing_tables)
        super(Neo4jPool, self).deactivate(address)

    def on_write_failure(self, address):
        """ Remove a writer address from the routing table, if present.
        """
        log.debug("[#0000]  C: <ROUTING> Removing writer %r", address)
        for database in self.routing_tables.keys():
            self.routing_tables[database].writers.discard(address)
        log.debug("[#0000]  C: <ROUTING> table=%r", self.routing_tables)


def _connect(resolved_address, timeout, keep_alive):
    """

    :param resolved_address:
    :param timeout: seconds
    :param keep_alive: True or False
    :return: socket object
    """

    s = None  # The socket

    try:
        if len(resolved_address) == 2:
            s = socket(AF_INET)
        elif len(resolved_address) == 4:
            s = socket(AF_INET6)
        else:
            raise ValueError("Unsupported address {!r}".format(resolved_address))
        t = s.gettimeout()
        if timeout:
            s.settimeout(timeout)
        log.debug("[#0000]  C: <OPEN> %s", resolved_address)
        s.connect(resolved_address)
        s.settimeout(t)
        keep_alive = 1 if keep_alive else 0
        s.setsockopt(SOL_SOCKET, SO_KEEPALIVE, keep_alive)
    except SocketTimeout:
        log.debug("[#0000]  C: <TIMEOUT> %s", resolved_address)
        log.debug("[#0000]  C: <CLOSE> %s", resolved_address)
        s.close()
        raise ServiceUnavailable("Timed out trying to establish connection to {!r}".format(resolved_address))
    except OSError as error:
        log.debug("[#0000]  C: <ERROR> %s %s", type(error).__name__,
                  " ".join(map(repr, error.args)))
        log.debug("[#0000]  C: <CLOSE> %s", resolved_address)
        s.close()
        raise ServiceUnavailable("Failed to establish connection to {!r} (reason {})".format(resolved_address, error))
    else:
        return s


def _secure(s, host, ssl_context):
    local_port = s.getsockname()[1]
    # Secure the connection if an SSL context has been provided
    if ssl_context:
        log.debug("[#%04X]  C: <SECURE> %s", local_port, host)
        try:
            sni_host = host if HAS_SNI and host else None
            s = ssl_context.wrap_socket(s, server_hostname=sni_host)
        except (SSLError, OSError) as cause:
            s.close()
            error = BoltSecurityError(message="Failed to establish encrypted connection.", address=(host, local_port))
            error.__cause__ = cause
            raise error
        else:
            # Check that the server provides a certificate
            der_encoded_server_certificate = s.getpeercert(binary_form=True)
            if der_encoded_server_certificate is None:
                s.close()
                raise BoltProtocolError("When using an encrypted socket, the server should always provide a certificate", address=(host, local_port))
    return s


def _handshake(s, resolved_address):
    """

    :param s: Socket
    :param resolved_address:

    :return: (socket, version, client_handshake, server_response_data)
    """
    local_port = s.getsockname()[1]

    # TODO: Optimize logging code
    handshake = Bolt.get_handshake()
    import struct
    handshake = struct.unpack(">16B", handshake)
    handshake = [handshake[i:i + 4] for i in range(0, len(handshake), 4)]

    supported_versions = [("0x%02X%02X%02X%02X" % (vx[0], vx[1], vx[2], vx[3])) for vx in handshake]

    log.debug("[#%04X]  C: <MAGIC> 0x%08X", local_port, int.from_bytes(Bolt.MAGIC_PREAMBLE, byteorder="big"))
    log.debug("[#%04X]  C: <HANDSHAKE> %s %s %s %s", local_port, *supported_versions)

    data = Bolt.MAGIC_PREAMBLE + Bolt.get_handshake()
    s.sendall(data)

    # Handle the handshake response
    ready_to_read = False
    while not ready_to_read:
        ready_to_read, _, _ = select((s,), (), (), 1)
    try:
        data = s.recv(4)
    except OSError:
        raise ServiceUnavailable("Failed to read any data from server {!r} "
                                 "after connected".format(resolved_address))
    data_size = len(data)
    if data_size == 0:
        # If no data is returned after a successful select
        # response, the server has closed the connection
        log.debug("[#%04X]  S: <CLOSE>", local_port)
        s.close()
        raise ServiceUnavailable("Connection to {address} closed without handshake response".format(address=resolved_address))
    if data_size != 4:
        # Some garbled data has been received
        log.debug("[#%04X]  S: @*#!", local_port)
        s.close()
        raise BoltProtocolError("Expected four byte Bolt handshake response from %r, received %r instead; check for incorrect port number" % (resolved_address, data), address=resolved_address)
    elif data == b"HTTP":
        log.debug("[#%04X]  S: <CLOSE>", local_port)
        s.close()
        raise ServiceUnavailable("Cannot to connect to Bolt service on {!r} "
                                 "(looks like HTTP)".format(resolved_address))
    agreed_version = data[-1], data[-2]
    log.debug("[#%04X]  S: <HANDSHAKE> 0x%06X%02X", local_port, agreed_version[1], agreed_version[0])
    return s, agreed_version, handshake, data


def connect(address, *, timeout, custom_resolver, ssl_context, keep_alive):
    """ Connect and perform a handshake and return a valid Connection object,
    assuming a protocol version can be agreed.
    """
    last_error = None
    # Establish a connection to the host and port specified
    # Catches refused connections see:
    # https://docs.python.org/2/library/errno.html
    log.debug("[#0000]  C: <RESOLVE> %s", address)

    for resolved_address in Address(address).resolve(resolver=custom_resolver):
        s = None
        try:
            host = address[0]
            s = _connect(resolved_address, timeout, keep_alive)
            s = _secure(s, host, ssl_context)
            return _handshake(s, address)
        except Exception as error:
            if s:
                s.close()
            last_error = error
    if last_error is None:
        raise ServiceUnavailable("Failed to resolve addresses for %s" % address)
    else:
        raise last_error



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j import (
    GraphDatabase,
    Neo4jDriver,
    TRUST_ALL_CERTIFICATES,
    TRUST_SYSTEM_CA_SIGNED_CERTIFICATES,
    DEFAULT_DATABASE,
)
from neo4j.api import (
    READ_ACCESS,
    WRITE_ACCESS,
)
from neo4j.exceptions import (
    ServiceUnavailable,
    ClientError,
    TransientError,
    SessionExpired,
    ConfigurationError,
)
from neo4j._exceptions import (
    BoltRoutingError,
    BoltSecurityError,
)
from tests.stub.conftest import StubCluster

# python -m pytest tests/stub/test_routingdriver.py -s -v


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/router.script",
        "v4x0/router.script",
    ]
)
def test_neo4j_uri_scheme_constructs_neo4j_driver(driver_info, test_script):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_neo4j_uri_scheme_constructs_neo4j_driver
    with StubCluster(test_script):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            assert isinstance(driver, Neo4jDriver)


@pytest.mark.skip(reason="Cant close the Stub Server gracefully")
@pytest.mark.parametrize(
    "test_script",
    [
        "v3/router.script",
        "v4x0/router.script",
    ]
)
def test_neo4j_uri_scheme_self_signed_certificate_constructs_neo4j_driver(driver_info, test_script):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_neo4j_uri_scheme_self_signed_certificate_constructs_neo4j_driver
    with StubCluster(test_script):
        uri = "neo4j+ssc://localhost:9001"

        test_config = {
            "user_agent": "test",
            "max_connection_lifetime": 1000,
            "max_connection_pool_size": 10,
            "keep_alive": False,
            "max_transaction_retry_time": 1,
            "resolver": None,
        }

        try:
            driver = GraphDatabase.driver(uri, auth=driver_info["auth_token"], **test_config)
            assert isinstance(driver, Neo4jDriver)
            driver.close()
        except ServiceUnavailable as error:
            assert isinstance(error.__cause__, BoltSecurityError)
            pytest.skip("Failed to establish encrypted connection")


@pytest.mark.skip(reason="Cant close the Stub Server gracefully")
@pytest.mark.parametrize(
    "test_script",
    [
        "v3/router.script",
        "v4x0/router.script",
    ]
)
def test_neo4j_uri_scheme_secure_constructs_neo4j_driver(driver_info, test_script):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_neo4j_uri_scheme_secure_constructs_neo4j_driver
    with StubCluster(test_script):
        uri = "neo4j+s://localhost:9001"

        test_config = {
            "user_agent": "test",
            "max_connection_lifetime": 1000,
            "max_connection_pool_size": 10,
            "keep_alive": False,
            "max_transaction_retry_time": 1,
            "resolver": None,
        }

        try:
            driver = GraphDatabase.driver(uri, auth=driver_info["auth_token"], **test_config)
            assert isinstance(driver, Neo4jDriver)
            driver.close()
        except ServiceUnavailable as error:
            assert isinstance(error.__cause__, BoltSecurityError)
            pytest.skip("Failed to establish encrypted connection")


@pytest.mark.parametrize(
    "test_uri",
    [
        "neo4j+ssc://localhost:9001",
        "neo4j+s://localhost:9001",
    ]
)
@pytest.mark.parametrize(
    "test_config, expected_failure, expected_failure_message",
    [
        ({"encrypted": False}, ConfigurationError, "The config settings"),
        ({"encrypted": True}, ConfigurationError, "The config settings"),
        ({"encrypted": True, "trust": TRUST_ALL_CERTIFICATES}, ConfigurationError, "The config settings"),
        ({"trust": TRUST_ALL_CERTIFICATES}, ConfigurationError, "The config settings"),
        ({"trust": TRUST_SYSTEM_CA_SIGNED_CERTIFICATES}, ConfigurationError, "The config settings"),
    ]
)
def test_neo4j_uri_scheme_secure_constructs_neo4j_driver_config_error(driver_info, test_uri, test_config, expected_failure, expected_failure_message):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_neo4j_uri_scheme_secure_constructs_neo4j_driver_config_error
    with pytest.raises(expected_failure) as error:
        driver = GraphDatabase.driver(test_uri, auth=driver_info["auth_token"], **test_config)

    assert error.match(expected_failure_message)


@pytest.mark.skip(reason="Flaky")
@pytest.mark.parametrize(
    "test_script",
    [
        "v3/router.script",
        "v4x0/router_port_9001_one_read_port_9004_one_write_port_9006.script",
    ]
)
def test_neo4j_driver_verify_connectivity(driver_info, test_script):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_neo4j_driver_verify_connectivity
    with StubCluster(test_script):
        driver = GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"], user_agent="test")
        assert isinstance(driver, Neo4jDriver)

    with StubCluster(test_script):
        assert driver.verify_connectivity() is not None
        driver.close()


@pytest.mark.skip(reason="Flaky")
@pytest.mark.parametrize(
    "test_script",
    [
        "v3/router.script",
        "v4x0/router_port_9001_one_read_port_9004_one_write_port_9006.script",
    ]
)
def test_neo4j_driver_verify_connectivity_server_down(driver_info, test_script):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_neo4j_driver_verify_connectivity_server_down
    with StubCluster(test_script):
        driver = GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"], user_agent="test")
        assert isinstance(driver, Neo4jDriver)

        with pytest.raises(ServiceUnavailable):
            driver.verify_connectivity()

        driver.close()


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/non_router.script",
        "v4x0/routing_table_failure_not_a_router.script",
    ]
)
def test_cannot_discover_servers_on_non_router(driver_info, test_script):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_cannot_discover_servers_on_non_router
    with StubCluster(test_script):
        with pytest.raises(ServiceUnavailable):
            with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]):
                pass


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/silent_router.script",
        "v4x0/routing_table_silent_router.script",
    ]
)
def test_cannot_discover_servers_on_silent_router(driver_info, test_script):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_cannot_discover_servers_on_silent_router
    with StubCluster(test_script):
        with pytest.raises(BoltRoutingError):
            with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]):
                pass


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/router.script",
        "v4x0/router.script",
    ]
)
def test_should_discover_servers_on_driver_construction(driver_info, test_script):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_should_discover_servers_on_driver_construction
    with StubCluster(test_script):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            table = driver._pool.routing_tables[DEFAULT_DATABASE]
            assert table.routers == {('127.0.0.1', 9001), ('127.0.0.1', 9002),
                                     ('127.0.0.1', 9003)}
            assert table.readers == {('127.0.0.1', 9004), ('127.0.0.1', 9005)}
            assert table.writers == {('127.0.0.1', 9006)}


@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/return_1.script"),
        ("v4x0/router.script", "v4x0/return_1_port_9004.script"),
    ]
)
def test_should_be_able_to_read(driver_info, test_scripts):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_should_be_able_to_read
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=READ_ACCESS, fetch_size=-1) as session:
                result = session.run("RETURN $x", {"x": 1})
                for record in result:
                    assert record["x"] == 1
                assert result.consume().server.address == ('127.0.0.1', 9004)


@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/create_a.script"),
        ("v4x0/router.script", "v4x0/create_test_node_port_9006.script"),
    ]
)
def test_should_be_able_to_write(driver_info, test_scripts):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_should_be_able_to_write
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=WRITE_ACCESS, fetch_size=-1) as session:
                result = session.run("CREATE (a $x)", {"x": {"name": "Alice"}})
                assert not list(result)
                assert result.consume().server.address == ('127.0.0.1', 9006)


@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/create_a.script"),
        ("v4x0/router.script", "v4x0/create_test_node_port_9006.script"),
    ]
)
def test_should_be_able_to_write_as_default(driver_info, test_scripts):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_should_be_able_to_write_as_default
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(fetch_size=-1) as session:
                result = session.run("CREATE (a $x)", {"x": {"name": "Alice"}})
                assert not list(result)
                assert result.consume().server.address == ('127.0.0.1', 9006)


@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/disconnect_on_run_9004.script"),
        ("v4x0/router.script", "v4x0/disconnect_on_run_port_9004.script"),
    ]
)
def test_routing_disconnect_on_run(driver_info, test_scripts):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_routing_disconnect_on_run
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with pytest.raises(SessionExpired):
                with driver.session(default_access_mode=READ_ACCESS, fetch_size=-1) as session:
                    session.run("RETURN $x", {"x": 1}).consume()


@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/disconnect_on_pull_all_9004.script"),
        ("v4x0/router.script", "v4x0/disconnect_on_pull_port_9004.script"),
    ]
)
def test_routing_disconnect_on_pull_all(driver_info, test_scripts):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_routing_disconnect_on_pull_all
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with pytest.raises(SessionExpired):
                with driver.session(default_access_mode=READ_ACCESS, fetch_size=-1) as session:
                    session.run("RETURN $x", {"x": 1}).consume()


@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/return_1.script"),
        ("v4x0/router.script", "v4x0/return_1_port_9004.script"),
    ]
)
def test_should_disconnect_after_fetching_autocommit_result(driver_info, test_scripts):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_should_disconnect_after_fetching_autocommit_result
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=READ_ACCESS, fetch_size=-1) as session:
                result = session.run("RETURN $x", {"x": 1})
                assert session._connection is not None
                result.consume()
                assert session._connection is None


@pytest.mark.parametrize(
    "test_scripts, test_run_args",
    [
        (("v3/router.script", "v3/return_1_twice_in_read_tx.script"), ("RETURN $x", {"x": 1})),
        (("v4x0/router.script", "v4x0/tx_return_1_twice_port_9004.script"), ("RETURN 1", )),
    ]
)
def test_should_disconnect_after_explicit_commit(driver_info, test_scripts, test_run_args):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_should_disconnect_after_explicit_commit
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=READ_ACCESS, fetch_size=-1) as session:
                with session.begin_transaction() as tx:
                    result = tx.run(*test_run_args)
                    assert session._connection is not None
                    result.consume()
                    assert session._connection is not None
                    result = tx.run(*test_run_args)
                    assert session._connection is not None
                    result.consume()
                    assert session._connection is not None
                assert session._connection is None


@pytest.mark.parametrize(
    "test_scripts, test_run_args",
    [
        (("v3/router.script", "v3/return_1_twice_in_read_tx.script"), ("RETURN $x", {"x": 1})),
        (("v4x0/router.script", "v4x0/tx_return_1_twice_port_9004.script"), ("RETURN 1", )),
    ]
)
def test_default_access_mode_defined_at_session_level(driver_info, test_scripts, test_run_args):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_default_access_mode_defined_at_session_level
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=READ_ACCESS, fetch_size=-1) as session:
                with session.begin_transaction() as tx:
                    result = tx.run(*test_run_args)
                    assert session._connection is not None
                    result.consume()
                    assert session._connection is not None
                    result = tx.run(*test_run_args)
                    assert session._connection is not None
                    result.consume()
                    assert session._connection is not None
                assert session._connection is None


@pytest.mark.parametrize(
    "test_scripts, test_run_args",
    [
        (("v3/router.script", "v3/return_1_twice.script"), ("RETURN $x", {"x": 1})),
        (("v4x0/router.script", "v4x0/return_1_twice_port_9004.script"), ("RETURN 1", )),
    ]
)
def test_should_reconnect_for_new_query(driver_info, test_scripts, test_run_args):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_should_reconnect_for_new_query
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=READ_ACCESS, fetch_size=-1) as session:
                result_1 = session.run(*test_run_args)
                assert session._connection is not None
                result_1.consume()
                assert session._connection is None
                result_2 = session.run(*test_run_args)
                assert session._connection is not None
                result_2.consume()
                assert session._connection is None


@pytest.mark.parametrize(
    "test_scripts, test_run_args",
    [
        (("v3/router.script", "v3/return_1_twice.script"), ("RETURN $x", {"x": 1})),
        (("v4x0/router.script", "v4x0/return_1_twice_port_9004.script"), ("RETURN 1", )),
    ]
)
def test_should_retain_connection_if_fetching_multiple_results(driver_info, test_scripts, test_run_args):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_should_retain_connection_if_fetching_multiple_results
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=READ_ACCESS, fetch_size=-1) as session:
                result_1 = session.run(*test_run_args)
                result_2 = session.run(*test_run_args)
                assert session._connection is not None
                result_1.consume()
                assert session._connection is not None
                result_2.consume()
                assert session._connection is None


@pytest.mark.parametrize(
    "test_scripts, test_run_args",
    [
        (("v3/router.script", "v3/return_1_four_times.script"), ("RETURN $x", {"x": 1})),
        (("v4x0/router.script", "v4x0/return_1_four_times_port_9004.script"), ("RETURN 1", )),
    ]
)
def test_two_sessions_can_share_a_connection(driver_info, test_scripts, test_run_args):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_two_sessions_can_share_a_connection
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            session_1 = driver.session(default_access_mode=READ_ACCESS, fetch_size=-1)
            session_2 = driver.session(default_access_mode=READ_ACCESS, fetch_size=-1)

            result_1a = session_1.run(*test_run_args)
            c = session_1._connection
            result_1a.consume()

            result_2a = session_2.run(*test_run_args)
            assert session_2._connection is c
            result_2a.consume()

            result_1b = session_1.run(*test_run_args)
            assert session_1._connection is c
            result_1b.consume()

            result_2b = session_2.run(*test_run_args)
            assert session_2._connection is c
            result_2b.consume()

            session_2.close()
            session_1.close()


@pytest.mark.parametrize(
    "test_scripts, test_run_args",
    [
        (("v3/get_routing_table.script", "v3/return_1_on_9002.script"), ("RETURN $x", {"x": 1})),
        (("v4x0/router_role_route_share_port_with_role_read_and_role_write.script", "v4x0/return_1_port_9002.script"), ("RETURN 1 AS x", )),
    ]
)
def test_should_call_get_routing_table_procedure(driver_info, test_scripts, test_run_args):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_should_call_get_routing_table_procedure
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=READ_ACCESS, fetch_size=-1) as session:
                result = session.run(*test_run_args)
                for record in result:
                    assert record["x"] == 1
                assert result.consume().server.address == ('127.0.0.1', 9002)


@pytest.mark.parametrize(
    "test_scripts, test_run_args",
    [
        (("v3/get_routing_table_with_context.script", "v3/return_1_on_9002.script"), ("RETURN $x", {"x": 1})),
        (("v4x0/router_get_routing_table_with_context.script", "v4x0/return_1_port_9002.script"), ("RETURN 1 AS x", )),
    ]
)
def test_should_call_get_routing_table_with_context(driver_info, test_scripts, test_run_args):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_should_call_get_routing_table_with_context
    with StubCluster(*test_scripts):
        uri = "neo4j://localhost:9001/?name=molly&age=1"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=READ_ACCESS, fetch_size=-1) as session:
                result = session.run(*test_run_args)
                for record in result:
                    assert record["x"] == 1
                assert result.consume().server.address == ('127.0.0.1', 9002)


@pytest.mark.parametrize(
    "test_scripts, test_run_args",
    [
        (("v3/router_no_writers.script", "v3/return_1_on_9005.script"), ("RETURN $x", {"x": 1})),
        (("v4x0/router_with_no_role_write.script", "v4x0/return_1_port_9005.script"), ("RETURN 1 AS x", )),
    ]
)
def test_should_serve_read_when_missing_writer(driver_info, test_scripts, test_run_args):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_should_serve_read_when_missing_writer
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=READ_ACCESS, fetch_size=-1) as session:
                result = session.run(*test_run_args)
                for record in result:
                    assert record["x"] == 1
                assert result.consume().server.address == ('127.0.0.1', 9005)


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/router_no_readers.script",
        "v4x0/router_with_no_role_read.script",
    ]
)
def test_should_error_when_missing_reader(driver_info, test_script):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_should_error_when_missing_reader
    with StubCluster(test_script):
        with pytest.raises(BoltRoutingError):
            GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"])


@pytest.mark.parametrize(
    "test_scripts, test_run_args",
    [
        (("v3/router.script", "v3/not_a_leader.script"), ("CREATE (n {name:'Bob'})", )),
        (("v4x0/router.script", "v4x0/run_with_failure_cluster_not_a_leader.script"), ("CREATE (n:TEST {name:'test'})", )),
    ]
)
def test_forgets_address_on_not_a_leader_error(driver_info, test_scripts, test_run_args):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_forgets_address_on_not_a_leader_error
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=WRITE_ACCESS, fetch_size=-1) as session:
                with pytest.raises(TransientError):
                    _ = session.run(*test_run_args)

                pool = driver._pool
                table = driver._pool.routing_tables[DEFAULT_DATABASE]

                # address might still have connections in the pool, failed instance just can't serve writes
                assert ('127.0.0.1', 9006) in pool.connections
                assert table.routers == {('127.0.0.1', 9001), ('127.0.0.1', 9002), ('127.0.0.1', 9003)}
                assert table.readers == {('127.0.0.1', 9004), ('127.0.0.1', 9005)}
                # writer 127.0.0.1:9006 should've been forgotten because of an error
                assert len(table.writers) == 0


@pytest.mark.parametrize(
    "test_scripts, test_run_args",
    [
        (("v3/router.script", "v3/forbidden_on_read_only_database.script"), ("CREATE (n {name:'Bob'})", )),
        (("v4x0/router.script", "v4x0/run_with_failure_forbidden_on_read_only_database.script"), ("CREATE (n:TEST {name:'test'})", )),
    ]
)
def test_forgets_address_on_forbidden_on_read_only_database_error(driver_info, test_scripts, test_run_args):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_forgets_address_on_forbidden_on_read_only_database_error
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=WRITE_ACCESS, fetch_size=-1) as session:
                with pytest.raises(TransientError):
                    _ = session.run(*test_run_args)

                pool = driver._pool
                table = driver._pool.routing_tables[DEFAULT_DATABASE]

                # address might still have connections in the pool, failed instance just can't serve writes
                assert ('127.0.0.1', 9006) in pool.connections
                assert table.routers == {('127.0.0.1', 9001), ('127.0.0.1', 9002), ('127.0.0.1', 9003)}
                assert table.readers == {('127.0.0.1', 9004), ('127.0.0.1', 9005)}
                # writer 127.0.0.1:9006 should've been forgotten because of an error
                assert len(table.writers) == 0


@pytest.mark.parametrize(
    "test_scripts, test_run_args",
    [
        (("v3/router.script", "v3/rude_reader.script"), ("RETURN 1", )),
        (("v4x0/router.script", "v4x0/disconnect_on_pull_port_9004.script"), ("RETURN $x", {"x": 1})),
    ]
)
def test_forgets_address_on_service_unavailable_error(driver_info, test_scripts, test_run_args):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_forgets_address_on_service_unavailable_error
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=READ_ACCESS, fetch_size=-1) as session:

                pool = driver._pool
                table = driver._pool.routing_tables[DEFAULT_DATABASE]
                table.readers.remove(('127.0.0.1', 9005))

                with pytest.raises(SessionExpired):
                    _ = session.run(*test_run_args)

                # address should have connections in the pool but be inactive, it has failed
                assert ('127.0.0.1', 9004) in pool.connections
                conns = pool.connections[('127.0.0.1', 9004)]
                conn = conns[0]
                assert conn._closed is True
                assert conn.in_use is True
                assert table.routers == {('127.0.0.1', 9001), ('127.0.0.1', 9002), ('127.0.0.1', 9003)}
                # reader 127.0.0.1:9004 should've been forgotten because of an error
                assert not table.readers
                assert table.writers == {('127.0.0.1', 9006)}

            assert conn.in_use is False


@pytest.mark.parametrize(
    "test_scripts, test_run_args",
    [
        (("v3/router.script", "v3/database_unavailable.script"), ("RETURN 1", )),
        (("v4x0/router.script", "v4x0/run_with_failure_database_unavailable.script"), ("RETURN 1", )),
    ]
)
def test_forgets_address_on_database_unavailable_error(driver_info, test_scripts, test_run_args):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_forgets_address_on_database_unavailable_error
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=READ_ACCESS, fetch_size=-1) as session:

                pool = driver._pool
                table = driver._pool.routing_tables[DEFAULT_DATABASE]
                table.readers.remove(('127.0.0.1', 9005))

                with pytest.raises(TransientError) as raised:
                    _ = session.run(*test_run_args)
                    assert raised.exception.title == "DatabaseUnavailable"

                pool = driver._pool
                table = driver._pool.routing_tables[DEFAULT_DATABASE]

                # address should not have connections in the pool, it has failed
                assert ('127.0.0.1', 9004) not in pool.connections
                assert table.routers == {('127.0.0.1', 9001), ('127.0.0.1', 9002), ('127.0.0.1', 9003)}
                # reader 127.0.0.1:9004 should've been forgotten because of an raised
                assert not table.readers
                assert table.writers == {('127.0.0.1', 9006)}


@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v4x1/router_get_routing_table_with_context.script", "v4x2/hello_with_routing_context_return_1_port_9002.script",),
        ("v4x1/router_get_routing_table_with_context.script", "v4x2/hello_with_routing_context_return_1_port_9002.script",),
    ]
)
def test_hello_routing(driver_info, test_scripts):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_hello_routing
    with StubCluster(*test_scripts):
        uri = "neo4j://localhost:9001/?region=china&policy=my_policy"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], user_agent="test") as driver:
            with driver.session(default_access_mode=READ_ACCESS, fetch_size=-1) as session:
                result = session.run("RETURN 1 AS x")
                for record in result:
                    assert record["x"] == 1
                address = result.consume().server.address
                assert address.host == "127.0.0.1"
                assert address.port == 9002



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

# tag::service-unavailable-import[]
from neo4j.exceptions import ServiceUnavailable
# end::service-unavailable-import[]


def service_unavailable_example(driver):
    with driver.session() as session:
        session.run("MATCH (_) DETACH DELETE _")

    # tag::service-unavailable[]
    def add_item():
        try:
            with driver.session() as session:
                session.write_transaction(lambda tx: tx.run("CREATE (a:Item)"))
            return True
        except ServiceUnavailable:
            return False
    # end::service-unavailable[]

    add_item()
    add_item()
    add_item()

    with driver.session() as session:
        items = session.run("MATCH (a:Item) RETURN count(a)").single().value()

    with driver.session() as session:
        session.run("MATCH (_) DETACH DELETE _")

    return items


def test_example():
    pytest.skip("Fix better error messages for the user. Be able to kill the server.")



# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from json import loads, dumps
from inspect import getmembers, isfunction
import testkitbackend.requests as requests


class Backend:
    def __init__(self, rd, wr):
        self._rd = rd
        self._wr = wr
        self.drivers = {}
        self.sessions = {}
        self.results = {}
        self.transactions = {}
        self.key = 0
        # Collect all request handlers
        self._requestHandlers = dict(
            [m for m in getmembers(requests, isfunction)])

    def next_key(self):
        self.key = self.key + 1
        return self.key

    def process_request(self):
        """ Reads next request from the stream and processes it.
        """
        in_request = False
        request = ""
        for line in self._rd:
            # Remove trailing newline
            line = line.decode('UTF-8').rstrip()
            if line == "#request begin":
                in_request = True
            elif line == "#request end":
                self._process(request)
                return True
            else:
                if in_request:
                    request = request + line
        return False

    def _process(self, request):
        """ Process a received request by retrieving handler that
        corresponds to the request name.
        """
        request = loads(request)
        if not isinstance(request, dict):
            raise Exception("Request is not an object")
        name = request.get('name', 'invalid')
        handler = self._requestHandlers.get(name)
        if not handler:
            raise Exception("No request handler for " + name)
        data = request["data"]
        handler(self, data)

    def send_response(self, name, data):
        """ Sends a response to backend.
        """
        response = {"name": name, "data": data}
        response = dumps(response)
        self._wr.write(b"#response begin\n")
        self._wr.write(bytes(response+"\n", "utf-8"))
        self._wr.write(b"#response end\n")
        self._wr.flush()
